from __future__ import annotations

import asyncio
import inspect
import logging
import socket
import time

# Import ValkeyStatClient with TYPE_CHECKING to avoid circular imports
from typing import (
    Any,
    Awaitable,
    Callable,
    Mapping,
    MutableMapping,
    Optional,
    Union,
    cast,
)

import redis.exceptions
import yarl
from glide import (
    AdvancedGlideClientConfiguration,
    GlideClient,
    GlideClientConfiguration,
    NodeAddress,
    ServerCredentials,
    TlsAdvancedConfiguration,
)
from redis.asyncio import ConnectionPool, Redis
from redis.asyncio.client import Pipeline
from redis.asyncio.sentinel import MasterNotFoundError, Sentinel, SlaveNotFoundError
from redis.backoff import ExponentialBackoff
from redis.retry import Retry

from ai.backend.common.utils import addr_to_hostport_pair
from ai.backend.logging import BraceStyleAdapter

from .types import RedisConnectionInfo, RedisHelperConfig, RedisTarget, ValkeyTarget
from .validators import DelimiterSeperatedList, HostPortPair

__all__ = (
    "execute",
    "get_redis_object",
)

_keepalive_options: MutableMapping[int, int] = {}


SSL_CERT_NONE = "none"
SSL_CERT_REQUIRED = "required"

# macOS does not support several TCP_ options
# so check if socket package includes TCP options before adding it
if (_TCP_KEEPIDLE := getattr(socket, "TCP_KEEPIDLE", None)) is not None:
    _keepalive_options[_TCP_KEEPIDLE] = 20

if (_TCP_KEEPINTVL := getattr(socket, "TCP_KEEPINTVL", None)) is not None:
    _keepalive_options[_TCP_KEEPINTVL] = 5

if (_TCP_KEEPCNT := getattr(socket, "TCP_KEEPCNT", None)) is not None:
    _keepalive_options[_TCP_KEEPCNT] = 3


_default_conn_opts: Mapping[str, Any] = {
    "socket_keepalive": True,
    "socket_keepalive_options": _keepalive_options,
    "retry": Retry(ExponentialBackoff(), 10),
    "retry_on_error": [
        redis.exceptions.ConnectionError,
        redis.exceptions.TimeoutError,
    ],
}
_default_conn_pool_opts: Mapping[str, Any] = {
    "max_connections": 16,
    # "timeout": 20.0,  # for redis-py 5.0+
}

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


# Leaving it as is since the CLI is still using redis-py version 4.x
async def execute(
    redis_obj: RedisConnectionInfo,
    func: Callable[[Union[Redis, Any]], Awaitable[Any]],
    *,
    service_name: Optional[str] = None,
    encoding: Optional[str] = None,
    command_timeout: Optional[float] = None,
) -> Any:
    """
    Executes a function that issues Redis commands or returns a pipeline/transaction of commands,
    with automatic retries upon temporary connection failures.

    Note that when retried, the given function may be executed *multiple* times, so the caller
    should take care of side-effects of it.
    """

    if not callable(func):
        raise TypeError("The func must be a function or a coroutinefunction with no arguments.")

    redis_client = redis_obj.client
    service_name = service_name or redis_obj.service_name
    reconnect_poll_interval = redis_obj.redis_helper_config.get("reconnect_poll_timeout", 0.0)

    first_trial = time.perf_counter()
    retry_log_count = 0
    last_log_time = first_trial

    def show_retry_warning(e: Exception, warn_on_first_attempt: bool = True) -> None:
        nonlocal retry_log_count, last_log_time
        now = time.perf_counter()
        if (warn_on_first_attempt and retry_log_count == 0) or now - last_log_time >= 10.0:
            log.warning(
                "Retrying due to interruption of Redis connection "
                "({}, conn-pool: {}, retrying-for: {:.3f}s)",
                repr(e),
                redis_obj.name,
                now - first_trial,
            )
            retry_log_count += 1
            last_log_time = now

    while True:
        try:
            async with redis_client:
                aw_or_pipe = func(redis_client)
                if isinstance(aw_or_pipe, Pipeline):
                    async with aw_or_pipe:
                        result = await aw_or_pipe.execute()
                elif inspect.isawaitable(aw_or_pipe):
                    result = await aw_or_pipe
                else:
                    raise ValueError(
                        "The redis execute's return value must be an awaitable"
                        " or redis.asyncio.client.Pipeline object"
                    )
                if isinstance(result, Pipeline):
                    # This happens when func is an async function that returns a pipeline.
                    async with result:
                        result = await result.execute()
                if encoding:
                    if isinstance(result, bytes):
                        return result.decode(encoding)
                    elif isinstance(result, dict):
                        newdict = {}
                        for k, v in result.items():
                            newdict[k.decode(encoding)] = v.decode(encoding)
                        return newdict
                else:
                    return result
        except (
            MasterNotFoundError,
            SlaveNotFoundError,
            redis.exceptions.ReadOnlyError,
            redis.exceptions.ConnectionError,
            ConnectionResetError,
            TypeError,  # 4.5.5 version of redis.py raises a TypeError when the Connection is closed.
        ) as e:
            warn_on_first_attempt = True
            if (
                isinstance(e, redis.exceptions.ConnectionError)
                and "Too many connections" in e.args[0]
            ):  # connection pool is full
                warn_on_first_attempt = False
            show_retry_warning(e, warn_on_first_attempt)
            await asyncio.sleep(reconnect_poll_interval)
            continue
        except (
            redis.exceptions.TimeoutError,
            asyncio.TimeoutError,
        ) as e:
            if command_timeout is not None:
                now = time.perf_counter()
                if now - first_trial >= command_timeout + 1.0:
                    show_retry_warning(e)
                first_trial = now
            continue
        except redis.exceptions.ResponseError as e:
            if "NOREPLICAS" in e.args[0]:
                show_retry_warning(e)
                await asyncio.sleep(reconnect_poll_interval)
                continue
            raise
        except asyncio.CancelledError:
            raise
        finally:
            await asyncio.sleep(0)


def _get_redis_url_schema(redis_target: RedisTarget) -> str:
    """
    Returns the Redis URL schema based on the Redis target configuration.
    """
    if redis_target.use_tls:
        return "rediss"
    return "redis"


def _parse_redis_url(redis_target: RedisTarget, db: int) -> yarl.URL:
    redis_url = redis_target.addr
    if redis_url is None:
        raise ValueError("Redis URL is not provided in the configuration.")

    schema = _get_redis_url_schema(redis_target)
    url = yarl.URL(f"{schema}://host").with_host(str(redis_url[0])).with_port(
        redis_url[1]
    ).with_password(redis_target.get("password")) / str(db)
    return url


def get_redis_object(
    redis_target: RedisTarget,
    *,
    name: str,
    db: int = 0,
    **kwargs,
) -> RedisConnectionInfo:
    redis_helper_config: RedisHelperConfig = cast(
        RedisHelperConfig, redis_target.redis_helper_config
    )
    conn_opts = {
        **_default_conn_opts,
        **kwargs,
        # "lib_name": None,  # disable implicit "CLIENT SETINFO" (for redis-py 5.0+)
        # "lib_version": None,  # disable implicit "CLIENT SETINFO" (for redis-py 5.0+)
    }
    conn_pool_opts = {
        **_default_conn_pool_opts,
    }
    if socket_timeout := redis_helper_config.get("socket_timeout"):
        conn_opts["socket_timeout"] = float(socket_timeout)
    if socket_connect_timeout := redis_helper_config.get("socket_connect_timeout"):
        conn_opts["socket_connect_timeout"] = float(socket_connect_timeout)
    if max_connections := redis_helper_config.get("max_connections"):
        conn_pool_opts["max_connections"] = int(max_connections)
    # for redis-py 5.0+
    # if connection_ready_timeout := redis_helper_config.get("connection_ready_timeout"):
    #     conn_pool_opts["timeout"] = float(connection_ready_timeout)
    if _sentinel_addresses := redis_target.get("sentinel"):
        sentinel_addresses: Any = None
        if isinstance(_sentinel_addresses, str):
            sentinel_addresses = DelimiterSeperatedList(HostPortPair).check_and_return(
                _sentinel_addresses
            )
        else:
            sentinel_addresses = _sentinel_addresses

        service_name = redis_target.get("service_name")
        password = redis_target.get("password")
        assert service_name is not None, (
            "config/redis/service_name is required when using Redis Sentinel"
        )

        kwargs = {
            "password": password,
            "ssl": redis_target.use_tls,
            "ssl_cert_reqs": SSL_CERT_NONE if redis_target.tls_skip_verify else SSL_CERT_REQUIRED,
        }
        sentinel = Sentinel(
            [(str(host), port) for host, port in sentinel_addresses],
            password=password,
            db=str(db),
            sentinel_kwargs={
                "password": password,
                **kwargs,
            },
        )
        return RedisConnectionInfo(
            client=sentinel.master_for(
                service_name=service_name,
                password=password,
                **conn_opts,
            ),
            sentinel=sentinel,
            name=name,
            service_name=service_name,
            redis_helper_config=redis_helper_config,
        )
    else:
        redis_url = redis_target.addr
        if redis_url is None:
            raise ValueError("Redis URL is not provided in the configuration.")

        url = _parse_redis_url(redis_target, db)
        return RedisConnectionInfo(
            # In redis-py 5.0.1+, we should migrate to `Redis.from_pool()` API
            client=Redis(
                connection_pool=ConnectionPool.from_url(
                    str(url),
                    **conn_pool_opts,
                    **conn_opts,
                ),
                **conn_opts,
                auto_close_connection_pool=True,
                ssl=redis_target.use_tls,
                ssl_cert_reqs=SSL_CERT_NONE if redis_target.tls_skip_verify else SSL_CERT_REQUIRED,
            ),
            sentinel=None,
            name=name,
            service_name=None,
            redis_helper_config=redis_helper_config,
        )


async def create_valkey_client(
    valkey_target: ValkeyTarget,
    *,
    name: str,
    db: int = 0,
    pubsub_channels: Optional[set[str]] = None,
) -> GlideClient:
    addresses: list[NodeAddress] = []
    if valkey_target.addr:
        host, port = addr_to_hostport_pair(valkey_target.addr)
        addresses.append(NodeAddress(host=str(host), port=int(port)))

    if sentinel_addresses := valkey_target.sentinel:
        for address in sentinel_addresses:
            host, port = addr_to_hostport_pair(address)
            addresses.append(NodeAddress(host=str(host), port=int(port)))

    credentials: Optional[ServerCredentials] = None
    if valkey_target.password:
        credentials = ServerCredentials(
            password=valkey_target.password,
        )
    pubsub_subscriptions: Optional[GlideClientConfiguration.PubSubSubscriptions] = None
    if pubsub_channels is not None:
        pubsub_subscriptions = GlideClientConfiguration.PubSubSubscriptions(
            channels_and_patterns={
                GlideClientConfiguration.PubSubChannelModes.Exact: pubsub_channels,
            },
            callback=None,
            context=None,
        )
    if not addresses:
        raise ValueError("At least one Redis address is required to create a GlideClient.")
    config = GlideClientConfiguration(
        addresses,
        use_tls=valkey_target.use_tls,
        advanced_config=AdvancedGlideClientConfiguration(
            tls_config=TlsAdvancedConfiguration(
                use_insecure_tls=valkey_target.tls_skip_verify,
            ),
        ),
        credentials=credentials,
        database_id=db,
        client_name=name,
        request_timeout=valkey_target.request_timeout or 1_000,  # default to 1 second
        pubsub_subscriptions=pubsub_subscriptions,
    )
    return await GlideClient.create(config)



import asyncio
import logging
import uuid
from typing import AsyncIterator, Optional

from ai.backend.common.events.fetcher import EventFetcher
from ai.backend.logging.utils import BraceStyleAdapter

from ...dispatcher import AbstractEvent
from ..hub import EventPropagator

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class WithCachePropagator(EventPropagator):
    """
    WithCachePropagator is an event propagator that handles broadcast events.
    It fetches the last event for a given task ID from the cache and propagates events
    to a queue for broadcasted events.
    It allows to prevent timing issues by ensuring that the last event is fetched before yielding new events.
    """

    _id: uuid.UUID
    _event_fetcher: EventFetcher
    _queue: asyncio.Queue[Optional[AbstractEvent]]
    _closed: bool = False

    def __init__(self, event_fetcher: EventFetcher) -> None:
        self._id = uuid.uuid4()
        self._event_fetcher = event_fetcher
        self._queue = asyncio.Queue()
        self._closed = False

    def id(self) -> uuid.UUID:
        """
        Get the unique identifier for the propagator.
        """
        return self._id

    async def _fetch_cached_event_safe(self, cache_id: str) -> Optional[AbstractEvent]:
        """
        Safely fetch cached event, handling any errors that may occur.
        Returns None if fetching fails or no cached event exists.
        """
        try:
            return await self._event_fetcher.fetch_cached_event(cache_id)
        except Exception as e:
            log.warning("Failed to fetch cached event for cache_id {}: {}", cache_id, e)
            return None

    async def receive(self, cache_id: str) -> AsyncIterator[AbstractEvent]:
        """
        First, it fetches the last event for a given cache ID using the event fetcher.
        If the last event is not None, it yields that event.
        Then, it enters a loop to receive events from the queue until closed,
        checking for cached events periodically.
        """
        # Initial check for cached event
        cached_event = await self._fetch_cached_event_safe(cache_id)
        if cached_event is not None:
            yield cached_event

        while not self._closed:
            try:
                # Try to get from queue with a 30 second timeout
                event = await asyncio.wait_for(self._queue.get(), timeout=30.0)
                if event is None:
                    break
                yield event
            except asyncio.TimeoutError:
                # On timeout, check for cached event again
                cached_event = await self._fetch_cached_event_safe(cache_id)
                if cached_event is not None:
                    yield cached_event
            except Exception as e:
                log.error("Error propagating event: {}", e)

    async def propagate_event(self, event: AbstractEvent) -> None:
        await self._queue.put(event)

    async def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        await self._queue.put(None)



import argparse
import ipaddress
import pathlib
from typing import Tuple, cast

from .types import HostPortPair


def port_no(s: str) -> int:
    try:
        port = int(s)
        assert port > 0
        assert port < 65536
    except (ValueError, AssertionError):
        msg = f"{s!r} is not a valid port number."
        raise argparse.ArgumentTypeError(msg)
    return port


def port_range(s: str) -> Tuple[int, int]:
    try:
        port_range = tuple(map(int, s.split("-")))
    except (TypeError, ValueError):
        msg = f"{s!r} should be a hyphen-separated pair of integers."
        raise argparse.ArgumentTypeError(msg)
    if len(port_range) != 2:
        msg = f"{s!r} should have exactly two integers."
        raise argparse.ArgumentTypeError(msg)
    if not (0 < port_range[0] < 65536):
        msg = f"{port_range[0]} is not a valid port number."
        raise argparse.ArgumentTypeError(msg)
    if not (0 < port_range[1] < 65536):
        msg = f"{port_range[1]} is not a valid port number."
        raise argparse.ArgumentTypeError(msg)
    if not (port_range[0] < port_range[1]):
        msg = f"{port_range[0]} should be less than {port_range[1]}."
        raise argparse.ArgumentTypeError(msg)
    return cast(Tuple[int, int], port_range)


def positive_int(s: str) -> int:
    try:
        val = int(s)
        assert val > 0
    except (ValueError, AssertionError):
        msg = f"{s!r} is not a positive integer."
        raise argparse.ArgumentTypeError(msg)
    return val


def non_negative_int(s: str) -> int:
    try:
        val = int(s)
        assert val >= 0
    except (ValueError, AssertionError):
        msg = f"{s!r} is not a non-negative integer."
        raise argparse.ArgumentTypeError(msg)
    return val


def host_port_pair(s: str) -> Tuple[ipaddress._BaseAddress, int]:
    host: str | ipaddress._BaseAddress
    pieces = s.rsplit(":", maxsplit=1)
    if len(pieces) == 1:
        msg = f"{s!r} should contain both IP address and port number."
        raise argparse.ArgumentTypeError(msg)
    elif len(pieces) == 2:
        # strip potential brackets in IPv6 hostname-port strings (RFC 3986).
        host = pieces[0].strip("[]")
        try:
            host = ipaddress.ip_address(host)
        except ValueError:
            # Let it be just a hostname.
            host = host
        try:
            port = int(pieces[1])
            assert port > 0
            assert port < 65536
        except (ValueError, AssertionError):
            msg = f"{pieces[1]!r} is not a valid port number."
            raise argparse.ArgumentTypeError(msg)
    return HostPortPair(host, port)


def ipaddr(s: str) -> ipaddress._BaseAddress:
    try:
        ip = ipaddress.ip_address(s.strip("[]"))
    except ValueError:
        msg = f"{s!r} is not a valid IP address."
        raise argparse.ArgumentTypeError(msg)
    return ip


def path(val: str) -> pathlib.Path:
    if val is None:
        return None
    p = pathlib.Path(val)
    if not p.exists():
        msg = f"{val!r} is not a valid file/dir path."
        raise argparse.ArgumentTypeError(msg)
    return p



from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import Any, Generic, Self, TypeVar

from ai.backend.common.types import DispatchResult

from ..reporter import ProgressReporter
from ..types import TaskName


class BaseBackgroundTaskArgs(ABC):
    @abstractmethod
    def to_metadata_body(self) -> dict[str, Any]:
        """
        Convert the arguments to a metadata body dictionary.
        This method should be implemented by subclasses to provide
        the specific conversion logic.
        """
        raise NotImplementedError("Subclasses must implement this method")

    @classmethod
    @abstractmethod
    def from_metadata_body(cls, body: Mapping[str, Any]) -> Self:
        """
        Create an instance from a metadata body dictionary.
        This method should be implemented by subclasses to provide
        the specific conversion logic.
        """
        raise NotImplementedError("Subclasses must implement this method")


TFunctionArgs = TypeVar("TFunctionArgs", bound=BaseBackgroundTaskArgs)


class BaseBackgroundTaskHandler(Generic[TFunctionArgs], ABC):
    @abstractmethod
    async def execute(self, reporter: ProgressReporter, args: TFunctionArgs) -> DispatchResult:
        """
        Execute the background task with the provided reporter and arguments.
        This method should be implemented by subclasses to provide
        the specific execution logic.
        """
        raise NotImplementedError("Subclasses must implement this method")

    @classmethod
    @abstractmethod
    def name(cls) -> TaskName:
        """
        Return the name of the background task.
        This method should be implemented by subclasses to provide
        the specific task name.
        """
        raise NotImplementedError("Subclasses must implement this method")

    @classmethod
    @abstractmethod
    def args_type(cls) -> type[TFunctionArgs]:
        """
        Return the type of arguments that this task expects.
        This method should be implemented by subclasses to provide
        the specific argument type.
        """
        raise NotImplementedError("Subclasses must implement this method")



import asyncio
import hashlib
import logging
import socket
from dataclasses import dataclass
from typing import Any, AsyncGenerator, Final, Mapping, Optional, Self

import glide
from aiotools.server import process_index

from ai.backend.common.clients.valkey_client.valkey_stream.client import ValkeyStreamClient
from ai.backend.common.defs import REDIS_STREAM_DB
from ai.backend.common.types import RedisTarget
from ai.backend.logging.utils import BraceStyleAdapter

from .queue import AbstractMessageQueue
from .types import BroadcastMessage, BroadcastPayload, MessageId, MQMessage

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

_DEFAULT_AUTOCLAIM_IDLE_TIMEOUT = 300_000  # 5 minutes
_DEFAULT_AUTOCLAIM_INTERVAL = 60_000
_DEFAULT_READ_BLOCK_MS: Final[int] = 10_000  # 10 second
_DEFAULT_READ_COUNT = 64


@dataclass
class RedisMQArgs:
    # Required arguments
    anycast_stream_key: str
    broadcast_channel: str
    consume_stream_keys: Optional[set[str]]
    subscribe_channels: Optional[set[str]]
    group_name: str
    node_id: str
    db: int
    # Optional arguments
    autoclaim_idle_timeout: int = _DEFAULT_AUTOCLAIM_IDLE_TIMEOUT
    autoclaim_start_id: Optional[str] = None


class RedisQueue(AbstractMessageQueue):
    _client: ValkeyStreamClient
    _consume_queue: asyncio.Queue[MQMessage]
    _subscribe_queue: asyncio.Queue[BroadcastMessage]
    _anycast_stream_key: str
    _broadcast_channel: str
    _group_name: str
    _consumer_id: str
    _redis_target: RedisTarget
    _closed: bool
    # loop tasks for consuming messages
    _loop_tasks: list[asyncio.Task]

    def __init__(
        self, client: ValkeyStreamClient, redis_target: RedisTarget, args: RedisMQArgs
    ) -> None:
        self._client = client
        self._consume_queue = asyncio.Queue()
        self._subscribe_queue = asyncio.Queue()
        self._anycast_stream_key = args.anycast_stream_key
        self._broadcast_channel = args.broadcast_channel
        self._group_name = args.group_name
        self._consumer_id = _generate_consumer_id(args.node_id)
        self._redis_target = redis_target
        self._closed = False
        start_id = args.autoclaim_start_id or "0-0"
        self._loop_tasks = []
        if args.consume_stream_keys:
            for stream_key in args.consume_stream_keys:
                self._loop_tasks.append(asyncio.create_task(self._read_messages_loop(stream_key)))
                self._loop_tasks.append(
                    asyncio.create_task(
                        self._auto_claim_loop(stream_key, start_id, args.autoclaim_idle_timeout)
                    )
                )
        if args.subscribe_channels:
            self._loop_tasks.append(asyncio.create_task(self._read_broadcast_messages_loop()))

    @classmethod
    async def create(cls, redis_target: RedisTarget, args: RedisMQArgs) -> Self:
        client = await ValkeyStreamClient.create(
            redis_target.to_valkey_target(),
            human_readable_name="event_producer.stream",
            db_id=args.db,
            pubsub_channels=args.subscribe_channels,
        )
        try:
            # Create consumer group if not exists
            await client.make_consumer_group(args.anycast_stream_key, args.group_name)
        except Exception:
            # Group may already exist
            pass
        return cls(client, redis_target, args)

    async def send(self, payload: dict[bytes, bytes]) -> None:
        """
        Send a message to the queue.
        If the queue is full, the oldest message will be removed.
        The new message will be added to the end of the queue.
        """
        if self._closed:
            raise RuntimeError("Queue is closed")
        await self._client.enqueue_stream_message(self._anycast_stream_key, payload)

    async def broadcast(self, payload: Mapping[str, Any]) -> None:
        """
        Broadcast a message to all subscribers.
        The message will be delivered to all subscribers.
        """
        if self._closed:
            raise RuntimeError("Queue is closed")
        await self._client.broadcast(self._broadcast_channel, payload)

    async def broadcast_with_cache(self, cache_id: str, payload: Mapping[str, str]) -> None:
        """
        Broadcast a message to all subscribers with cache.
        The message will be delivered to all subscribers.
        """
        if self._closed:
            raise RuntimeError("Queue is closed")
        await self._client.broadcast_with_cache(self._broadcast_channel, cache_id, payload)

    async def fetch_cached_broadcast_message(self, cache_id: str) -> Optional[Mapping[str, str]]:
        """
        Fetch a cached broadcast message by cache_id.
        This method retrieves the cached message from the broadcast channel.
        """
        if self._closed:
            raise RuntimeError("Queue is closed")
        return await self._client.fetch_cached_broadcast_message(cache_id)

    async def broadcast_batch(self, events: list[BroadcastPayload]) -> None:
        """
        Broadcast multiple messages in a batch with optional caching.
        This method broadcasts multiple messages to all subscribers.
        """
        if self._closed:
            raise RuntimeError("Queue is closed")
        await self._client.broadcast_batch(self._broadcast_channel, events)

    async def consume_queue(self) -> AsyncGenerator[MQMessage, None]:  # type: ignore
        """
        Consume messages from the queue.
        This method will block until a message is available.

        This is a normal queue, so the message will be delivered to one consumer.
        Messages are consumed only once by one consumer.

        Consumer should call `done` method to acknowledge the message when it is processed.
        If the consumer does not call `done`, the message will be re-delivered after the
        `autoclaim_idle_timeout` period.
        """
        while not self._closed:
            try:
                yield await self._consume_queue.get()
            except asyncio.CancelledError:
                break

    async def subscribe_queue(self) -> AsyncGenerator[BroadcastMessage, None]:  # type: ignore
        while not self._closed:
            try:
                yield await self._subscribe_queue.get()
            except asyncio.CancelledError:
                break

    async def done(self, msg_id: MessageId) -> None:
        await self._done(self._anycast_stream_key, msg_id)

    async def _done(self, stream_key: str, msg_id: MessageId) -> None:
        await self._client.done_stream_message(stream_key, self._group_name, msg_id)

    async def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        for task in self._loop_tasks:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                log.debug("Task {} cancelled", task.get_name())
        await self._client.close()

    async def _auto_claim_loop(
        self, stream_key: str, autoclaim_start_id: str, autoclaim_idle_timeout: int
    ) -> None:
        log.debug("Starting auto claim loop for stream {}", stream_key)
        while not self._closed:
            try:
                next_start_id, claimed = await self._auto_claim(
                    stream_key, autoclaim_start_id, autoclaim_idle_timeout
                )
                if claimed:
                    autoclaim_start_id = next_start_id
                    continue
            except glide.TimeoutError:
                # If the auto claim times out, we just continue to the next iteration
                pass
            except glide.ClosingError:
                log.info(
                    "Client connection closed, stopping auto claim loop for stream {}", stream_key
                )
                break
            except glide.GlideError as e:
                await self._failover_consumer(stream_key, e)
            except Exception as e:
                log.exception("Error while auto claiming messages: {}", e)
            await asyncio.sleep(_DEFAULT_AUTOCLAIM_INTERVAL / 1000)

    async def _auto_claim(
        self, stream_key: str, autoclaim_start_id: str, autoclaim_idle_timeout: int
    ) -> tuple[str, bool]:
        message = await self._client.auto_claim_stream_message(
            stream_key,
            self._group_name,
            self._consumer_id,
            autoclaim_start_id,
            autoclaim_idle_timeout,
        )
        if message is None:
            return autoclaim_start_id, False
        for msg in message.messages:
            mq_msg = MQMessage(msg.msg_id, {**msg.payload})
            if mq_msg.retry():
                await self._retry_message(stream_key, mq_msg)
                continue
            # discard the message
            await self._done(stream_key, mq_msg.msg_id)
        return autoclaim_start_id, len(message.messages) > 0

    async def _retry_message(self, stream_key: str, message: MQMessage) -> None:
        await self._client.reque_stream_message(
            stream_key, self._group_name, message.msg_id, message.payload
        )

    async def _read_messages_loop(self, stream_key: str) -> None:
        log.info("Starting read messages loop for stream {}", stream_key)
        target = self._redis_target.to_valkey_target()
        # Set the request timeout to be longer than the read block time
        target.request_timeout = (_DEFAULT_READ_BLOCK_MS // 1000) + 1  # add 1 second buffer
        client = await ValkeyStreamClient.create(
            target,
            human_readable_name="event_producer.stream",
            db_id=REDIS_STREAM_DB,
        )
        while not self._closed:
            try:
                await self._read_messages(client, stream_key)
            except glide.ClosingError:
                log.info(
                    "Client connection closed, stopping read messages loop for stream {}",
                    stream_key,
                )
                break
            except glide.GlideError as e:
                await self._failover_consumer(stream_key, e)
            except Exception as e:
                log.error("Error while reading messages: {}", e)

    async def _read_messages(self, client: ValkeyStreamClient, stream_key: str) -> None:
        payload = await client.read_consumer_group(
            stream_key,
            self._group_name,
            self._consumer_id,
            count=_DEFAULT_READ_COUNT,
            block_ms=_DEFAULT_READ_BLOCK_MS,
        )
        if not payload:
            return
        for msg in payload:
            mq_msg = MQMessage(msg_id=msg.msg_id, payload={**msg.payload})
            await self._consume_queue.put(mq_msg)

    async def _read_broadcast_messages_loop(self) -> None:
        log.info("Starting read broadcast messages loop")
        while not self._closed:
            try:
                await self._read_broadcast_messages()
            except glide.ClosingError:
                log.info("Client connection closed, stopping read broadcast messages loop")
                break
            except Exception as e:
                log.error("Error while reading broadcast messages: {}", e)

    async def _read_broadcast_messages(self) -> None:
        payload = await self._client.receive_broadcast_message()
        msg = BroadcastMessage(payload)
        await self._subscribe_queue.put(msg)

    async def _failover_consumer(self, stream_key: str, e: Exception) -> None:
        # If the group does not exist, create it
        # and start the auto claim loop again
        if "NOGROUP" in str(e):
            log.warning(
                "Consumer group does not exist. Creating group {} for stream {}",
                self._group_name,
                stream_key,
            )
            try:
                await self._client.make_consumer_group(
                    stream_key,
                    self._group_name,
                )
            except Exception as internal_exception:
                log.exception(
                    "Error while creating consumer group {} for stream {}: {}",
                    self._group_name,
                    stream_key,
                    internal_exception,
                )
        else:
            log.exception("Error while reading messages: {}", e)


def _generate_consumer_id(node_id: Optional[str]) -> str:
    h = hashlib.sha1()
    h.update(str(node_id or socket.getfqdn()).encode("utf8"))
    hostname_hash = h.hexdigest()
    h = hashlib.sha1()
    h.update(__file__.encode("utf8"))
    installation_path_hash = h.hexdigest()
    pidx = process_index.get(0)
    return f"{hostname_hash}:{installation_path_hash}:{pidx}"



import json
import uuid
from dataclasses import dataclass
from typing import Any, Self, Sequence

from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.json import dump_json_str
from ai.backend.common.types import ValkeyTarget

from ..service_discovery import ServiceDiscovery, ServiceMetadata

_DEFAULT_PREFIX = "service_discovery"
_DEFAULT_TTL = 60 * 3  # 3 minutes


@dataclass
class RedisServiceDiscoveryArgs:
    valkey_target: ValkeyTarget
    db_id: int = 0
    ttl: int = _DEFAULT_TTL  # 3 minutes
    prefix: str = _DEFAULT_PREFIX


class RedisServiceDiscovery(ServiceDiscovery):
    _valkey_client: ValkeyLiveClient
    _ttl: int
    _prefix: str

    def __init__(self, valkey_client: ValkeyLiveClient, args: RedisServiceDiscoveryArgs) -> None:
        self._valkey_target = args.valkey_target
        self._db_id = args.db_id
        self._ttl = args.ttl
        self._prefix = args.prefix
        self._valkey_client = valkey_client

    @classmethod
    async def create(cls, args: RedisServiceDiscoveryArgs) -> Self:
        valkey_client = await ValkeyLiveClient.create(
            args.valkey_target,
            db_id=args.db_id,
            human_readable_name="service_discovery",
        )
        return cls(valkey_client, args)

    async def close(self) -> None:
        """Close the ValkeyLiveClient connection."""
        await self._valkey_client.close()

    async def _hsetex(self, key: str, mapping: dict[str, Any]) -> None:
        """Set hash fields with expiry using ValkeyLiveClient."""
        # Convert values to JSON strings
        json_mapping = {k: dump_json_str(v) for k, v in mapping.items()}
        await self._valkey_client.hset_with_expiry(key, json_mapping, self._ttl)

    async def _scan_keys(self, pattern: str) -> list[str]:
        """Scan keys matching pattern using ValkeyLiveClient."""
        return await self._valkey_client.scan_keys(pattern)

    async def _hget_json(self, name: str) -> ServiceMetadata:
        """Get hash fields and parse JSON values using ValkeyLiveClient."""
        raw_hash = await self._valkey_client.hgetall_str(name)
        if not raw_hash:
            raise ValueError(f"Service key {name} not found.")

        result: dict[str, Any] = {}
        for key, value in raw_hash.items():
            result[key] = json.loads(value)

        return ServiceMetadata.from_dict(result)

    async def register(self, service_meta: ServiceMetadata) -> None:
        key = self._service_prefix(service_meta.service_group, service_meta.id)
        await self._hsetex(key, service_meta.to_dict())

    async def unregister(self, service_group: str, service_id: uuid.UUID) -> None:
        """Unregister a service using ValkeyLiveClient."""
        key = self._service_prefix(service_group, service_id)
        await self._valkey_client.delete_key(key)

    async def heartbeat(self, service_meta: ServiceMetadata) -> None:
        service_meta.health_status.update_heartbeat()
        key = self._service_prefix(service_meta.service_group, service_meta.id)
        await self._hsetex(key, service_meta.to_dict())

    async def discover(self) -> Sequence[ServiceMetadata]:
        pattern = f"{self._prefix}.*.*"
        keys = await self._scan_keys(pattern)
        if not keys:
            raise ValueError("No service groups found.")
        return [await self._hget_json(key) for key in keys]

    async def get_service_group(self, service_group: str) -> Sequence[ServiceMetadata]:
        pattern = f"{self._service_group_prefix(service_group)}.*"
        keys = await self._scan_keys(pattern)
        if not keys:
            raise ValueError(f"No services found in group {service_group}.")
        return [await self._hget_json(key) for key in keys]

    async def get_service(self, service_group: str, service_id: uuid.UUID) -> ServiceMetadata:
        key = self._service_prefix(service_group, service_id)
        return await self._hget_json(key)

    def _service_group_prefix(self, service_group: str) -> str:
        return f"{self._prefix}.{service_group}"

    def _service_prefix(self, service_group: str, service_id: uuid.UUID) -> str:
        return f"{self._service_group_prefix(service_group)}.{service_id}"



from abc import ABC, abstractmethod
from dataclasses import dataclass

from .types import AbstractEvent


@dataclass
class PrepareEventReportArgs:
    pass


@dataclass
class CompleteEventReportArgs:
    duration: float


class AbstractEventReporter(ABC):
    @abstractmethod
    async def prepare_event_report(
        self,
        event: AbstractEvent,
        arg: PrepareEventReportArgs,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    async def complete_event_report(
        self,
        event: AbstractEvent,
        arg: CompleteEventReportArgs,
    ) -> None:
        raise NotImplementedError



import base64
from contextlib import ExitStack, contextmanager
from dataclasses import dataclass
from typing import Iterator, Mapping, Optional, Self, cast

from ai.backend.common import msgpack
from ai.backend.common.contexts.request_id import with_request_id
from ai.backend.common.contexts.user import with_user
from ai.backend.common.data.user.types import UserData
from ai.backend.common.json import dump_json, load_json
from ai.backend.logging.utils import with_log_context_fields

_DEFAULT_RETRY_FIELD = b"_retry_count"
_DEFAULT_MAX_RETRIES = 3

type MessageId = bytes


@dataclass
class BroadcastPayload:
    """Payload data for broadcasting with optional cache."""

    payload: Mapping[str, str]
    cache_id: Optional[str] = None


@dataclass
class BroadcastMessage:
    payload: Mapping[str, str]


@dataclass
class MQMessage:
    msg_id: MessageId
    payload: dict[bytes, bytes]

    def retry(self) -> bool:
        """
        Retry the message.
        If the message has been retried more than the maximum number of retries,
        the message will be discarded.
        The retry count is stored in the message payload.
        """
        if self._retry_count() > _DEFAULT_MAX_RETRIES:
            return False
        self.payload[_DEFAULT_RETRY_FIELD] = str(self._retry_count() + 1).encode("utf-8")
        return True

    def _retry_count(self) -> int:
        """
        Get the retry count of the message.
        The retry count is the number of times the message has been re-delivered.
        """
        return int(self.payload.get(_DEFAULT_RETRY_FIELD, b"0"))


@dataclass
class MessageMetadata:
    request_id: Optional[str] = None
    user: Optional[UserData] = None

    def serialize(self) -> bytes:
        """
        Serialize the metadata to bytes.
        """
        return dump_json(self)

    @classmethod
    def deserialize(cls, data: str | bytes) -> Self:
        """
        Deserialize the metadata from bytes.
        """
        result = load_json(data)
        if "user_id" in result:
            del result["user_id"]
        if "user" in result:
            user_data = result["user"]
            if isinstance(user_data, dict):
                user = UserData(**user_data)
                result["user"] = user
            else:
                result["user"] = None
        return cls(**result)

    @contextmanager
    def apply_context(self) -> Iterator[None]:
        """
        Context manager to apply all context variables stored in metadata.
        """
        with ExitStack() as stack:
            log_fields: dict[str, str] = {}
            if self.request_id:
                stack.enter_context(with_request_id(self.request_id))
                log_fields["request_id"] = self.request_id
            if self.user:
                stack.enter_context(with_user(self.user))
                log_fields["user_id"] = str(self.user.user_id)
            if log_fields:
                stack.enter_context(with_log_context_fields(log_fields))
            yield


@dataclass
class MessagePayload:
    name: str
    source: str
    args: tuple[bytes, ...]
    metadata: Optional[MessageMetadata] = None

    def serialize_anycast(self) -> dict[bytes, bytes]:
        """
        Serialize the message payload to a dictionary.
        """
        result = {
            b"name": self.name.encode(),
            b"source": self.source.encode(),
            b"args": msgpack.packb(self.args),
        }
        if self.metadata:
            result[b"metadata"] = self.metadata.serialize()
        return result

    def serialize_broadcast(self) -> dict[str, str]:
        """
        Serialize the message payload to a dictionary.
        The keys are bytes and the values are bytes.
        """
        args = base64.b64encode(msgpack.packb(self.args)).decode("ascii")
        result = {
            "name": self.name,
            "source": self.source,
            "args": args,
        }
        if self.metadata:
            result["metadata"] = self.metadata.serialize().decode("utf-8")
        return result

    @classmethod
    def from_anycast(cls, payload: Mapping[bytes, bytes]) -> Self:
        """
        Deserialize the message payload from a dictionary.
        The keys are bytes and the values are bytes.
        """
        metadata = None
        if b"metadata" in payload:
            metadata = MessageMetadata.deserialize(payload[b"metadata"])
        return cls(
            name=payload[b"name"].decode("utf-8"),
            source=payload[b"source"].decode("utf-8"),
            args=cast(tuple[bytes, ...], msgpack.unpackb(payload[b"args"])),
            metadata=metadata,
        )

    @classmethod
    def from_broadcast(cls, payload: Mapping[str, str]) -> Self:
        """
        Deserialize the message payload from a dictionary.
        The keys are bytes and the values are bytes.
        """
        args_bytes = base64.b64decode(payload["args"])
        metadata = None
        if "metadata" in payload:
            metadata = MessageMetadata.deserialize(payload["metadata"])
        return cls(
            name=payload["name"],
            source=payload["source"],
            args=cast(tuple[bytes, ...], msgpack.unpackb(args_bytes)),
            metadata=metadata,
        )



from abc import ABC, abstractmethod
from typing import Any, Mapping, Optional


class UserEvent(ABC):
    @abstractmethod
    def event_name(self) -> Optional[str]:
        """Get the name of the event."""
        raise NotImplementedError

    @abstractmethod
    def retry_count(self) -> Optional[int]:
        """Get the retry count for the event."""
        raise NotImplementedError

    @abstractmethod
    def user_event_mapping(self) -> Mapping[str, Any]:
        """Get the user event as a dictionary."""
        raise NotImplementedError

    @abstractmethod
    def is_close_event(self) -> bool:
        """
        Indicate if the event is a close event.
        Close events are used to signal that the event is no longer needed to be propagated.
        """
        raise NotImplementedError



import asyncio
import enum
import os
from typing import Optional, Self

import psutil
from prometheus_client import Counter, Gauge, Histogram, generate_latest

from ai.backend.common.exception import BackendAIError, ErrorCode


class APIMetricObserver:
    _instance: Optional[Self] = None

    _request_count: Counter
    _request_duration_sec: Histogram

    def __init__(self) -> None:
        self._request_count = Counter(
            name="backendai_api_request_count",
            documentation="Total number of API requests",
            labelnames=["method", "endpoint", "domain", "operation", "error_detail", "status_code"],
        )
        self._request_duration_sec = Histogram(
            name="backendai_api_request_duration_sec",
            documentation="Duration of API requests in milliseconds",
            labelnames=["method", "endpoint", "domain", "operation", "error_detail", "status_code"],
            buckets=[0.001, 0.01, 0.1, 0.5, 1, 2, 5, 10, 30],
        )

    @classmethod
    def instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def _inc_request_total(
        self, *, method: str, endpoint: str, error_code: Optional[ErrorCode], status_code: int
    ) -> None:
        self._request_count.labels(
            method=method,
            endpoint=endpoint,
            domain=error_code.domain if error_code else "",
            operation=error_code.operation if error_code else "",
            error_detail=error_code.error_detail if error_code else "",
            status_code=status_code,
        ).inc()

    def _observe_request_duration(
        self,
        *,
        method: str,
        endpoint: str,
        error_code: Optional[ErrorCode],
        status_code: int,
        duration: float,
    ) -> None:
        self._request_duration_sec.labels(
            method=method,
            endpoint=endpoint,
            domain=error_code.domain if error_code else "",
            operation=error_code.operation if error_code else "",
            error_detail=error_code.error_detail if error_code else "",
            status_code=status_code,
        ).observe(duration)

    def observe_request(
        self,
        *,
        method: str,
        endpoint: str,
        error_code: Optional[ErrorCode],
        status_code: int,
        duration: float,
    ) -> None:
        self._inc_request_total(
            method=method, endpoint=endpoint, error_code=error_code, status_code=status_code
        )
        self._observe_request_duration(
            method=method,
            endpoint=endpoint,
            status_code=status_code,
            error_code=error_code,
            duration=duration,
        )


class GraphQLMetricObserver:
    _instance: Optional[Self] = None

    _request_count: Counter
    _request_duration_sec: Histogram

    def __init__(self) -> None:
        self._request_count = Counter(
            name="backendai_graphql_request_count",
            documentation="Total number of API requests",
            labelnames=[
                "operation_type",
                "field_name",
                "parent_type",
                "operation_name",
                "domain",
                "operation",
                "error_detail",
                "success",
            ],
        )
        self._request_duration_sec = Histogram(
            name="backendai_graphql_request_duration_sec",
            documentation="Duration of API requests in milliseconds",
            labelnames=[
                "operation_type",
                "field_name",
                "parent_type",
                "operation_name",
                "domain",
                "operation",
                "error_detail",
                "success",
            ],
            buckets=[0.001, 0.01, 0.1, 0.5, 1, 2, 5, 10, 30],
        )

    @classmethod
    def instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def _inc_request_total(
        self,
        *,
        operation_type: str,
        field_name: str,
        parent_type: str,
        operation_name: str,
        error_code: Optional[ErrorCode],
        success: bool,
    ) -> None:
        self._request_count.labels(
            operation_type=operation_type,
            field_name=field_name,
            parent_type=parent_type,
            operation_name=operation_name,
            domain=error_code.domain if error_code else "",
            operation=error_code.operation if error_code else "",
            error_detail=error_code.error_detail if error_code else "",
            success=success,
        ).inc()

    def _observe_request_duration(
        self,
        *,
        operation_type: str,
        field_name: str,
        parent_type: str,
        operation_name: str,
        error_code: Optional[ErrorCode],
        success: bool,
        duration: float,
    ) -> None:
        self._request_duration_sec.labels(
            operation_type=operation_type,
            field_name=field_name,
            parent_type=parent_type,
            operation_name=operation_name,
            domain=error_code.domain if error_code else "",
            operation=error_code.operation if error_code else "",
            error_detail=error_code.error_detail if error_code else "",
            success=success,
        ).observe(duration)

    def observe_request(
        self,
        *,
        operation_type: str,
        field_name: str,
        parent_type: str,
        operation_name: str,
        error_code: Optional[ErrorCode],
        success: bool,
        duration: float,
    ) -> None:
        self._inc_request_total(
            operation_type=operation_type,
            field_name=field_name,
            parent_type=parent_type,
            operation_name=operation_name,
            error_code=error_code,
            success=success,
        )
        self._observe_request_duration(
            operation_type=operation_type,
            field_name=field_name,
            parent_type=parent_type,
            operation_name=operation_name,
            error_code=error_code,
            success=success,
            duration=duration,
        )


class EventMetricObserver:
    _instance: Optional[Self] = None

    _event_count: Counter
    _event_failure_count: Counter
    _event_processing_time_sec: Histogram

    def __init__(self) -> None:
        self._event_count = Counter(
            name="backendai_event_count",
            documentation="Total number of events processed",
            labelnames=["event_type"],
        )
        self._event_failure_count = Counter(
            name="backendai_event_failure_count",
            documentation="Number of failed events",
            labelnames=["event_type", "exception", "domain", "operation", "error_detail"],
        )
        self._event_processing_time_sec = Histogram(
            name="backendai_event_processing_time_sec",
            documentation="Processing time of events in seconds",
            labelnames=["event_type", "status", "domain", "operation", "error_detail"],
            buckets=[0.001, 0.01, 0.1, 0.5, 1, 2, 5, 10, 30],
        )

    @classmethod
    def instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def observe_event_success(self, *, event_type: str, duration: float) -> None:
        self._event_count.labels(event_type=event_type).inc()
        self._event_processing_time_sec.labels(
            event_type=event_type,
            status="success",
            domain="",
            operation="",
            error_detail="",
        ).observe(duration)

    def observe_event_failure(
        self,
        *,
        event_type: str,
        duration: float,
        exception: BaseException,
    ) -> None:
        error_code: ErrorCode
        if isinstance(exception, BackendAIError):
            error_code = exception.error_code()
        else:
            error_code = ErrorCode.default()
        exception_name = exception.__class__.__name__
        self._event_failure_count.labels(
            event_type=event_type,
            exception=exception_name,
            domain=error_code.domain,
            operation=error_code.operation,
            error_detail=error_code.error_detail,
        ).inc()
        self._event_count.labels(event_type=event_type).inc()
        self._event_processing_time_sec.labels(
            event_type=event_type,
            status="failure",
            domain=error_code.domain,
            operation=error_code.operation,
            error_detail=error_code.error_detail,
        ).observe(duration)


class BgTaskMetricObserver:
    _instance: Optional[Self] = None

    _bgtask_count: Gauge
    _bgtask_done_count: Counter
    _bgtask_processing_time: Histogram

    def __init__(self) -> None:
        self._bgtask_count = Gauge(
            name="backendai_bgtask_count",
            documentation="Total number of background tasks processed",
            labelnames=["task_name"],
        )
        self._bgtask_done_count = Counter(
            name="backendai_bgtask_done_count",
            documentation="Number of completed background tasks",
            labelnames=["task_name", "status", "domain", "operation", "error_detail"],
        )
        self._bgtask_processing_time = Histogram(
            name="backendai_bgtask_processing_time_sec",
            documentation="Processing time of background tasks in seconds",
            labelnames=["task_name", "status", "domain", "operation", "error_detail"],
            buckets=[0.1, 1, 10, 30, 60, 300, 600],
        )

    @classmethod
    def instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def observe_bgtask_started(self, *, task_name: str) -> None:
        self._bgtask_count.labels(task_name=task_name).inc()

    def observe_bgtask_done(
        self, *, task_name: str, status: str, duration: float, error_code: Optional[ErrorCode]
    ) -> None:
        self._bgtask_count.labels(task_name=task_name).dec()
        self._bgtask_processing_time.labels(
            task_name=task_name,
            status=status,
            domain=error_code.domain if error_code else "",
            operation=error_code.operation if error_code else "",
            error_detail=error_code.error_detail if error_code else "",
        ).observe(duration)
        self._bgtask_done_count.labels(
            task_name=task_name,
            status=status,
            domain=error_code.domain if error_code else "",
            operation=error_code.operation if error_code else "",
            error_detail=error_code.error_detail if error_code else "",
        ).inc()


class ActionMetricObserver:
    _instance: Optional[Self] = None

    _action_count: Counter
    _action_duration_sec: Histogram

    def __init__(self) -> None:
        self._action_count = Counter(
            name="backendai_action_count",
            documentation="Total number of actions",
            labelnames=[
                "entity_type",
                "operation_type",
                "status",
                "domain",
                "operation",
                "error_detail",
            ],
        )
        self._action_duration_sec = Histogram(
            name="backendai_action_duration_sec",
            documentation="Duration of actions in seconds",
            labelnames=[
                "entity_type",
                "operation_type",
                "status",
                "domain",
                "operation",
                "error_detail",
            ],
            buckets=[0.001, 0.01, 0.1, 0.5, 1, 2, 5, 10, 30],
        )

    @classmethod
    def instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def observe_action(
        self,
        *,
        entity_type: str,
        operation_type: str,
        status: str,
        duration: float,
        error_code: Optional[ErrorCode],
    ) -> None:
        self._action_count.labels(
            entity_type=entity_type,
            operation_type=operation_type,
            status=status,
            domain=error_code.domain if error_code else "",
            operation=error_code.operation if error_code else "",
            error_detail=error_code.error_detail if error_code else "",
        ).inc()
        self._action_duration_sec.labels(
            entity_type=entity_type,
            operation_type=operation_type,
            status=status,
            domain=error_code.domain if error_code else "",
            operation=error_code.operation if error_code else "",
            error_detail=error_code.error_detail if error_code else "",
        ).observe(duration)


class DomainType(enum.StrEnum):
    VALKEY = "valkey"
    REPOSITORY = "repository"
    CLIENT = "client"


class LayerType(enum.StrEnum):
    # Repository layers
    AGENT = "agent"
    AUTH = "auth"
    ARTIFACT = "artifact"
    ARTIFACT_REGISTRY = "artifact_registry"
    CONTAINER_REGISTRY = "container_registry"
    DOMAIN = "domain"
    GROUP = "group"
    IMAGE = "image"
    KEYPAIR_RESOURCE_POLICY = "keypair_resource_policy"
    METRIC = "metric"
    MODEL_SERVING = "model_serving"
    PROJECT_RESOURCE_POLICY = "project_resource_policy"
    RESOURCE_PRESET = "resource_preset"
    SCHEDULE = "schedule"
    SESSION = "session"
    USER = "user"
    USER_RESOURCE_POLICY = "user_resource_policy"
    VFOLDER = "vfolder"
    PERMISSION_CONTROL = "permission_control"
    OBJECT_STORAGE = "object_storage"

    # Valkey client layers
    VALKEY_CONTAINER_LOG = "valkey_container_log"
    VALKEY_IMAGE = "valkey_image"
    VALKEY_LIVE = "valkey_live"
    VALKEY_RATE_LIMIT = "valkey_rate_limit"
    VALKEY_SCHEDULE = "valkey_schedule"
    VALKEY_SESSION = "valkey_session"
    VALKEY_STAT = "valkey_stat"
    VALKEY_STREAM = "valkey_stream"
    VALKEY_BGTASK = "valkey_bgtask"

    # Client layers
    AGENT_CLIENT = "agent_client"
    STORAGE_PROXY_CLIENT = "storage_proxy_client"
    WSPROXY_CLIENT = "wsproxy_client"


# Backward compatibility
ClientType = DomainType


class LayerMetricObserver:
    _instance: Optional[Self] = None

    _layer_operation_triggered_count: Gauge
    _layer_operation_count: Counter
    _layer_retry_count: Counter
    _layer_operation_duration_sec: Histogram

    def __init__(self) -> None:
        self._layer_operation_triggered_count = Gauge(
            name="backendai_layer_operation_triggered_count",
            documentation="Number of layer operations triggered",
            labelnames=["domain", "layer", "operation"],
        )
        self._layer_operation_count = Counter(
            name="backendai_layer_operation_count",
            documentation="Total number of layer operations",
            labelnames=["domain", "layer", "operation", "success"],
        )
        self._layer_retry_count = Counter(
            name="backendai_layer_retry_count",
            documentation="Number of retries for layer operations",
            labelnames=["domain", "layer", "operation"],
        )
        self._layer_operation_duration_sec = Histogram(
            name="backendai_layer_operation_duration_sec",
            documentation="Duration of layer operations in seconds",
            labelnames=["domain", "layer", "operation", "success"],
            buckets=[0.001, 0.01, 0.1, 0.5, 1, 2, 5, 10, 30],
        )

    @classmethod
    def instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def observe_layer_operation_triggered(
        self,
        *,
        domain: DomainType,
        layer: LayerType,
        operation: str,
    ) -> None:
        self._layer_operation_triggered_count.labels(
            domain=domain,
            layer=layer,
            operation=operation,
        ).inc()

    def observe_layer_retry(
        self,
        *,
        domain: DomainType,
        layer: LayerType,
        operation: str,
    ) -> None:
        self._layer_retry_count.labels(
            domain=domain,
            layer=layer,
            operation=operation,
        ).inc()

    def observe_layer_operation(
        self,
        *,
        domain: DomainType,
        layer: LayerType,
        operation: str,
        success: bool,
        duration: float,
    ) -> None:
        self._layer_operation_triggered_count.labels(
            domain=domain,
            layer=layer,
            operation=operation,
        ).dec()  # Decrement the triggered count since the operation is now complete
        self._layer_operation_count.labels(
            domain=domain,
            layer=layer,
            operation=operation,
            success=str(success),
        ).inc()
        self._layer_operation_duration_sec.labels(
            domain=domain,
            layer=layer,
            operation=operation,
            success=str(success),
        ).observe(duration)


class SystemMetricObserver:
    _instance: Optional[Self] = None

    _async_task_count: Gauge
    _cpu_usage_percent: Gauge
    _memory_used_rss: Gauge
    _memory_used_vms: Gauge

    def __init__(self) -> None:
        self._async_task_count = Gauge(
            name="backendai_async_task_count",
            documentation="Number of active async tasks",
        )
        self._cpu_usage_percent = Gauge(
            name="backendai_cpu_usage_percent",
            documentation="CPU usage of the process",
        )
        self._memory_used_rss = Gauge(
            name="backendai_memory_used_rss",
            documentation="Memory used by the process in RSS",
        )
        self._memory_used_vms = Gauge(
            name="backendai_memory_used_vms",
            documentation="Memory used by the process in VMS",
        )

    @classmethod
    def instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def observe(self) -> None:
        self._async_task_count.set(len(asyncio.all_tasks()))
        proc = psutil.Process(os.getpid())
        self._cpu_usage_percent.set(proc.cpu_percent())
        self._memory_used_rss.set(proc.memory_info().rss)
        self._memory_used_vms.set(proc.memory_info().vms)


class SweeperMetricObserver:
    _instance: Optional[Self] = None

    _session_sweep_count: Counter
    _kernel_sweep_count: Counter

    def __init__(self) -> None:
        self._session_sweep_count = Counter(
            name="backendai_sweep_session_count",
            documentation="Total number of session sweeps",
            labelnames=["status", "success"],
        )
        self._kernel_sweep_count = Counter(
            name="backendai_sweep_kernel_count",
            documentation="Total number of kernel sweeps",
            labelnames=["success"],
        )

    @classmethod
    def instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def observe_session_sweep(self, *, status: str, success: bool) -> None:
        self._session_sweep_count.labels(status=status, success=success).inc()

    def observe_kernel_sweep(self, *, success: bool) -> None:
        self._kernel_sweep_count.labels(success=success).inc()


class EventPropagatorMetricObserver:
    _instance: Optional[Self] = None

    _propagator_count: Gauge
    _propagator_alias_count: Gauge
    _propagator_registration_count: Counter
    _propagator_unregistration_count: Counter

    def __init__(self) -> None:
        self._propagator_count = Gauge(
            name="backendai_event_propagator_count",
            documentation="Current number of active event propagators",
        )
        self._propagator_alias_count = Gauge(
            name="backendai_event_propagator_alias_count",
            documentation="Current number of event propagator aliases",
            labelnames=["domain", "alias_id"],
        )
        self._propagator_registration_count = Counter(
            name="backendai_event_propagator_registration_count",
            documentation="Total number of event propagator registrations",
        )
        self._propagator_unregistration_count = Counter(
            name="backendai_event_propagator_unregistration_count",
            documentation="Total number of event propagator unregistrations",
        )

    @classmethod
    def instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def observe_propagator_registered(self, *, aliases: list[tuple[str, str]]) -> None:
        self._propagator_count.inc()
        self._propagator_registration_count.inc()
        for domain, alias_id in aliases:
            self._propagator_alias_count.labels(domain=domain, alias_id=alias_id).inc()

    def observe_propagator_unregistered(self, *, aliases: list[tuple[str, str]]) -> None:
        self._propagator_count.dec()
        self._propagator_unregistration_count.inc()
        for domain, alias_id in aliases:
            self._propagator_alias_count.labels(domain=domain, alias_id=alias_id).dec()


class CommonMetricRegistry:
    _instance: Optional[Self] = None

    api: APIMetricObserver
    gql: GraphQLMetricObserver
    event: EventMetricObserver
    bgtask: BgTaskMetricObserver
    system: SystemMetricObserver
    sweeper: SweeperMetricObserver
    event_propagator_observer: EventPropagatorMetricObserver

    def __init__(self) -> None:
        self.api = APIMetricObserver.instance()
        self.gql = GraphQLMetricObserver.instance()
        self.event = EventMetricObserver.instance()
        self.bgtask = BgTaskMetricObserver.instance()
        self.system = SystemMetricObserver.instance()
        self.sweeper = SweeperMetricObserver.instance()
        self.event_propagator_observer = EventPropagatorMetricObserver.instance()

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def to_prometheus(self) -> str:
        self.system.observe()
        return generate_latest().decode("utf-8")


class StageObserver:
    _instance: Optional[Self] = None

    _stage_count: Counter

    def __init__(self) -> None:
        self._stage_count = Counter(
            name="backendai_stage_count",
            documentation="Count stage occurrences",
            labelnames=["stage", "upper_layer"],
        )

    @classmethod
    def instance(cls) -> Self:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def observe_stage(self, *, stage: str, upper_layer: str) -> None:
        self._stage_count.labels(stage=stage, upper_layer=upper_layer).inc()



import asyncio
import logging
from collections.abc import Sequence

from ai.backend.logging.utils import BraceStyleAdapter

from ..observer.types import AbstractObserver
from ..resource.types import AbstractResource

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class Runner:
    """
    Runner is a utility class that manages the lifecycle of resources and observers.
    It sets up resources, starts observers, and cleans up resources when closed.
    Parameters
    ----------
    resources : Sequence[AbstractResource]
        A sequence of resources to be managed by the runner.
        Each resource should implement the AbstractResource interface.

    Examples
    --------
    runner = Runner(resources=[MyResource(), AnotherResource()])
    await runner.register_observer(MyObserver())
    await runner.start()
    """

    _resources: Sequence[AbstractResource]
    _closed_event: asyncio.Event

    def __init__(self, resources: Sequence[AbstractResource]):
        self._resources = resources
        self._closed_event = asyncio.Event()

    async def register_observer(self, observer: AbstractObserver) -> None:
        """
        Create a task to run the observer.
        This will run the observer in a loop until the runner is closed.
        """
        if self._closed_event.is_set():
            raise RuntimeError("Runner is already closed.")
        log.info("Starting observer: {}", observer.name)
        asyncio.create_task(self._run_observer(observer))

    async def _run_observer(self, observer: AbstractObserver) -> None:
        while not self._closed_event.is_set():
            try:
                async with asyncio.timeout(observer.timeout()):
                    await observer.observe()
            except asyncio.TimeoutError:
                log.warning(
                    "Observer {} timed out after {} seconds.",
                    observer.name,
                    observer.timeout(),
                )
            except Exception as e:
                log.exception(
                    "Error while observing: {}",
                    e,
                )
            await asyncio.sleep(observer.observe_interval())
        await observer.cleanup()
        log.info(
            "Observer closed: {}",
            observer.name,
        )

    async def _setup(self) -> None:
        for resource in self._resources:
            try:
                await resource.setup()
                log.info(
                    "Resource setup: {}",
                    resource.name,
                )
            except Exception as e:
                log.exception(
                    "Error while setting up resource: {}",
                    e,
                )
                raise

    async def _cleanup(self) -> None:
        for resource in self._resources:
            try:
                await resource.release()
                log.info(
                    "Resource released: {}",
                    resource.name,
                )
            except Exception as e:
                log.exception(
                    "Error while releasing resource: {}",
                    e,
                )

    async def start(self) -> None:
        """
        Start the runner.
        This will setup all resources and start runner loop.
        It will run until the runner is closed.
        """
        if self._closed_event.is_set():
            raise RuntimeError("Runner is already closed.")
        log.info("Starting runner.")
        try:
            await self._setup()
        except Exception as e:
            log.exception(
                "Error while starting runner: {}",
                e,
            )
            await self._cleanup()
            raise
        asyncio.create_task(self._run())
        log.info("Runner started.")

    async def _run(self) -> None:
        try:
            await self._closed_event.wait()
        finally:
            log.info("cleaning up runner.")
            await self._cleanup()
            log.info("Runner closed.")

    async def close(self) -> None:
        """
        Close the runner.
        This will stop all observers and cleanup all resources.
        """
        self._closed_event.set()
        # Give the event loop a chance to schedule the _run() task
        # that's waiting on the closed_event
        await asyncio.sleep(0)



import uuid
from dataclasses import dataclass
from typing import Optional, Self, override

from ai.backend.common.events.event_types.kernel.types import KernelLifecycleEventReason
from ai.backend.common.events.types import (
    AbstractBroadcastEvent,
    EventCacheDomain,
    EventDomain,
)
from ai.backend.common.events.user_event.user_event import UserEvent
from ai.backend.common.types import SessionId


@dataclass
class BaseSessionEvent(AbstractBroadcastEvent):
    session_id: SessionId

    @classmethod
    @override
    def event_domain(cls) -> EventDomain:
        return EventDomain.SESSION

    @override
    def domain_id(self) -> Optional[str]:
        return str(self.session_id)

    @override
    def user_event(self) -> Optional[UserEvent]:
        return None


@dataclass
class DoTerminateSessionEvent(BaseSessionEvent):
    reason: KernelLifecycleEventReason

    def serialize(self) -> tuple:
        return (
            str(self.session_id),
            self.reason,
        )

    @classmethod
    @override
    def deserialize(cls, value: tuple) -> Self:
        return cls(
            SessionId(uuid.UUID(value[0])),
            value[1],
        )

    @classmethod
    @override
    def event_name(cls) -> str:
        return "do_terminate_session"


@dataclass
class SessionCreationEvent(BaseSessionEvent):
    creation_id: str
    reason: KernelLifecycleEventReason = KernelLifecycleEventReason.UNKNOWN

    @override
    def serialize(self) -> tuple:
        return (
            str(self.session_id),
            self.creation_id,
            self.reason,
        )

    @classmethod
    @override
    def deserialize(cls, value: tuple):
        return cls(
            SessionId(uuid.UUID(value[0])),
            value[1],
            value[2],
        )

    @override
    def user_event(self) -> Optional[UserEvent]:
        return None


@dataclass
class SessionEnqueuedBroadcastEvent(SessionCreationEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "session_enqueued"


@dataclass
class SessionScheduledBroadcastEvent(SessionCreationEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "session_scheduled"


@dataclass
class SessionCheckingPrecondBroadcastEvent(SessionCreationEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "session_checking_precondition"


@dataclass
class SessionPreparingBroadcastEvent(SessionCreationEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "session_preparing"


@dataclass
class SessionCancelledBroadcastEvent(SessionCreationEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "session_cancelled"


@dataclass
class SessionStartedBroadcastEvent(SessionCreationEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "session_started"


@dataclass
class SessionTerminationEvent(BaseSessionEvent):
    reason: str = ""

    @override
    def serialize(self) -> tuple:
        return (
            str(self.session_id),
            self.reason,
        )

    @classmethod
    @override
    def deserialize(cls, value: tuple):
        return cls(
            SessionId(uuid.UUID(value[0])),
            value[1],
        )

    @override
    def user_event(self) -> Optional[UserEvent]:
        return None


@dataclass
class SessionTerminatingBroadcastEvent(SessionTerminationEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "session_terminating"


@dataclass
class SessionTerminatedBroadcastEvent(SessionTerminationEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "session_terminated"


@dataclass
class SessionResultEvent(BaseSessionEvent):
    reason: KernelLifecycleEventReason = KernelLifecycleEventReason.UNKNOWN
    exit_code: int = -1

    def serialize(self) -> tuple:
        return (
            str(self.session_id),
            self.reason,
            self.exit_code,
        )

    @classmethod
    def deserialize(cls, value: tuple):
        return cls(
            SessionId(uuid.UUID(value[0])),
            value[1],
            value[2],
        )

    @override
    def user_event(self) -> Optional[UserEvent]:
        return None


@dataclass
class SessionSuccessBroadcastEvent(SessionResultEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "session_success"


@dataclass
class SessionFailureBroadcastEvent(SessionResultEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "session_failure"


@dataclass
class SessionSchedulingEventData:
    """Data for each session in batch scheduling event."""

    session_id: SessionId
    creation_id: str


@dataclass
class SchedulingBroadcastEvent(AbstractBroadcastEvent):
    """Individual scheduling event for a session status transition."""

    session_id: SessionId
    creation_id: str
    status_transition: str  # "SCHEDULED", "PREPARING", "CREATING", etc.
    reason: str  # "self-terminated", "user-requested", etc.

    @classmethod
    @override
    def event_domain(cls) -> EventDomain:
        return EventDomain.SESSION

    @override
    def domain_id(self) -> Optional[str]:
        return str(self.session_id)

    @override
    def serialize(self) -> tuple:
        return (
            str(self.session_id),
            self.creation_id,
            self.status_transition,
            self.reason,
        )

    @classmethod
    @override
    def deserialize(cls, value: tuple) -> Self:
        return cls(
            session_id=SessionId(uuid.UUID(value[0])),
            creation_id=value[1],
            status_transition=value[2],
            reason=value[3],
        )

    @classmethod
    @override
    def event_name(cls) -> str:
        return "scheduling"

    @override
    def user_event(self) -> Optional[UserEvent]:
        return None

    @classmethod
    @override
    def cache_domain(cls) -> Optional[EventCacheDomain]:
        return EventCacheDomain.SESSION_SCHEDULER



"""Exceptions for leader election module."""

from __future__ import annotations

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class LeaderElectionError(BackendAIError):
    """Base exception for leader election errors."""

    error_type = "https://api.backend.ai/probs/leader-election-error"
    error_title = "Leader Election Error"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.LEADER_ELECTION,
            operation=ErrorOperation.EXECUTE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class AlreadyStartedError(LeaderElectionError):
    """Raised when trying to start an already started election or register tasks after start."""

    error_type = "https://api.backend.ai/probs/leader-already-started"
    error_title = "Leader Election Already Started"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.LEADER_ELECTION,
            operation=ErrorOperation.START,
            error_detail=ErrorDetail.ALREADY_EXISTS,
        )



import asyncio
import asyncio.staggered
import base64
import enum
import functools
import json
import logging
import os
import socket
import uuid
from ipaddress import _BaseAddress as BaseIPAddress
from ipaddress import _BaseNetwork as BaseIPNetwork
from ipaddress import ip_address
from pathlib import Path, PosixPath
from typing import Awaitable, Callable, Iterable, Optional

import aiodns
import aiohttp
import ifaddr
import psutil

from .utils import curl

__all__ = (
    "detect_cloud",
    "current_provider",
    "get_instance_id",
    "get_instance_ip",
    "get_instance_type",
    "get_instance_region",
    "get_root_fs_type",
    "get_wsl_version",
)

log = logging.getLogger(__spec__.name)


class CloudProvider(enum.StrEnum):
    AWS = "amazon"
    GCP = "google"
    AZURE = "azure"


def is_containerized() -> bool:
    """
    Check if I am running inside a Linux container.
    """
    try:
        cginfo = Path("/proc/self/cgroup").read_text()
        if "/docker/" in cginfo or "/lxc/" in cginfo:
            return True
        return False
    except IOError:
        return False


async def _detect_aws(session: aiohttp.ClientSession) -> CloudProvider:
    async with session.get(
        "http://169.254.169.254/latest/meta-data/",
    ):
        return CloudProvider.AWS


async def _detect_azure(session: aiohttp.ClientSession) -> CloudProvider:
    async with session.get(
        "http://169.254.169.254/metadata/instance/compute",
        params={"api-version": "2021-02-01"},
        headers={"Metadata": "true"},
    ):
        return CloudProvider.AZURE


async def _detect_gcp(session: aiohttp.ClientSession) -> CloudProvider:
    async with session.get(
        "http://169.254.169.254/computeMetadata/v1/instance/id",
        headers={"Metadata-Flavor": "Google"},
    ):
        return CloudProvider.GCP


async def detect_cloud() -> Optional[CloudProvider]:
    """
    Detect the cloud provider using asyncio.staggered_race()
    to get the fastest returning result from multiple metadata URL detectors.
    """
    async with aiohttp.ClientSession(
        raise_for_status=True,
        timeout=aiohttp.ClientTimeout(connect=0.3),
    ) as session:
        detection_tasks = [
            functools.partial(_detect_aws, session),
            functools.partial(_detect_azure, session),
            functools.partial(_detect_gcp, session),
        ]
        winner_value, winner_index, exceptions = await asyncio.staggered.staggered_race(
            detection_tasks,
            delay=0.001,
        )
        if winner_value is not None:
            return winner_value
    return None


def fetch_local_ipaddrs(cidr: BaseIPNetwork) -> Iterable[BaseIPAddress]:
    proto = socket.AF_INET if cidr.version == 4 else socket.AF_INET6
    for adapter in ifaddr.get_adapters():
        if not adapter.ips:
            continue
        for entry in adapter.ips:
            if entry.is_IPv4 and proto == socket.AF_INET:
                assert isinstance(entry.ip, str)
                addr = ip_address(entry.ip)
            elif entry.is_IPv6 and proto == socket.AF_INET6:
                assert isinstance(entry.ip, tuple)
                addr = ip_address(entry.ip[0])
            else:
                continue
            if addr in cidr:
                yield addr


def get_root_fs_type() -> tuple[PosixPath, str]:
    for partition in psutil.disk_partitions():
        if partition.mountpoint == "/":
            return PosixPath(partition.device), partition.fstype
    raise RuntimeError("Could not find the root filesystem from the mounts.")


def get_wsl_version() -> int:
    """
    Returns the current WSL version we are running on, and 0 if we are not on WSL.

    ref) https://github.com/snapcore/snapd/blob/3a88dc38ca122eba97192dba3aad30f3bd3e3081/release/release.go#L116-L172
    """
    if not Path("/proc/sys/fs/binfmt_misc/WSLInterop").exists() and not Path("/run/WSL").exists():
        return 0
    try:
        _, root_fs_type = get_root_fs_type()
    except RuntimeError:
        return 2
    if root_fs_type in ("wslfs", "lxfs"):
        return 1
    return 2


# Detect upon module load.
try:
    try:
        loop = asyncio.get_running_loop()
        current_provider = loop.run_until_complete(detect_cloud())
    except RuntimeError:
        current_provider = asyncio.run(detect_cloud())
except Exception as e:
    log.warning(f"Failed to detect cloud provider: {e}")
    current_provider = None

if current_provider is None:
    log.info("Detected environment: on-premise setup")
    log.info("The agent node ID is set using the hostname.")
else:
    log.info(f"Detected environment: {current_provider} cloud")
    log.info("The agent node ID will follow the instance ID.")

_defined: bool = False
get_instance_id: Callable[[], Awaitable[str]]
get_instance_ip: Callable[[Optional[BaseIPNetwork]], Awaitable[str]]
get_instance_type: Callable[[], Awaitable[str]]
get_instance_region: Callable[[], Awaitable[str]]


def _define_functions():
    global _defined
    global get_instance_id
    global get_instance_ip
    global get_instance_type
    global get_instance_region
    if _defined:
        return

    match current_provider:
        case CloudProvider.AWS:
            # ref: http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/ec2-instance-metadata.html
            _metadata_prefix = "http://169.254.169.254/latest/meta-data/"
            _dynamic_prefix = "http://169.254.169.254/latest/dynamic/"

            async def _get_instance_id() -> str:
                return await curl(
                    _metadata_prefix + "instance-id", lambda: f"i-{socket.gethostname()}"
                )

            async def _get_instance_ip(subnet_hint: Optional[BaseIPNetwork] = None) -> str:
                return await curl(_metadata_prefix + "local-ipv4", "127.0.0.1")

            async def _get_instance_type() -> str:
                return await curl(_metadata_prefix + "instance-type", "unknown")

            async def _get_instance_region() -> str:
                doc = await curl(_dynamic_prefix + "instance-identity/document", None)
                if doc is None:
                    return "amazon/unknown"
                region = json.loads(doc)["region"]
                return f"amazon/{region}"

        case CloudProvider.AZURE:
            # ref: https://learn.microsoft.com/en-us/azure/virtual-machines/instance-metadata-serviceb
            _metadata_prefix = "http://169.254.169.254/metadata/instance"

            async def _get_instance_id() -> str:
                data = await curl(
                    _metadata_prefix,
                    None,
                    params={"api-version": "2021-02-01"},
                    headers={"Metadata": "true"},
                )
                if data is None:
                    return f"i-{socket.gethostname()}"
                o = json.loads(data)
                vm_name = o["compute"]["name"]  # unique within the resource group
                vm_id = uuid.UUID(o["compute"]["vmId"])  # prevent conflicts across resource group
                vm_id_hash = base64.b32encode(vm_id.bytes[-5:]).decode().lower()
                return f"i-{vm_name}-{vm_id_hash}"

            async def _get_instance_ip(subnet_hint: Optional[BaseIPNetwork] = None) -> str:
                data = await curl(
                    _metadata_prefix,
                    None,
                    params={"api-version": "2021-02-01"},
                    headers={"Metadata": "true"},
                )
                if data is None:
                    return "127.0.0.1"
                o = json.loads(data)
                return o["network"]["interface"][0]["ipv4"]["ipAddress"][0]["privateIpAddress"]

            async def _get_instance_type() -> str:
                data = await curl(
                    _metadata_prefix,
                    None,
                    params={"api-version": "2021-02-01"},
                    headers={"Metadata": "true"},
                )
                if data is None:
                    return "unknown"
                o = json.loads(data)
                return o["compute"]["vmSize"]

            async def _get_instance_region() -> str:
                data = await curl(
                    _metadata_prefix,
                    None,
                    params={"api-version": "2021-02-01"},
                    headers={"Metadata": "true"},
                )
                if data is None:
                    return "azure/unknown"
                o = json.loads(data)
                region = o["compute"]["location"]
                return f"azure/{region}"

        case CloudProvider.GCP:
            # ref: https://cloud.google.com/compute/docs/storing-retrieving-metadata
            _metadata_prefix = "http://169.254.169.254/computeMetadata/v1/"

            async def _get_instance_id() -> str:
                vm_id = await curl(
                    _metadata_prefix + "instance/id",
                    None,
                    headers={"Metadata-Flavor": "Google"},
                )
                if vm_id is None:
                    return f"i-{socket.gethostname()}"
                vm_name = await curl(
                    _metadata_prefix + "instance/name",
                    None,
                    headers={"Metadata-Flavor": "Google"},
                )
                vm_id_hash = base64.b32encode(int(vm_id).to_bytes(8, "big")[-5:]).decode().lower()
                return f"i-{vm_name}-{vm_id_hash}"

            async def _get_instance_ip(subnet_hint: Optional[BaseIPNetwork] = None) -> str:
                return await curl(
                    _metadata_prefix + "instance/network-interfaces/0/ip",
                    "127.0.0.1",
                    headers={"Metadata-Flavor": "Google"},
                )

            async def _get_instance_type() -> str:
                value = await curl(
                    _metadata_prefix + "instance/machine-type",
                    "unknown",
                    headers={"Metadata-Flavor": "Google"},
                )
                return value.rsplit("/")[-1]

            async def _get_instance_region() -> str:
                value = await curl(
                    _metadata_prefix + "instance/zone",
                    "unknown",
                    headers={"Metadata-Flavor": "Google"},
                )
                region = value.rsplit("/")[-1]
                return f"google/{region}"

        case _:
            _metadata_prefix = None

            async def _get_instance_id() -> str:
                return f"i-{socket.gethostname()}"

            async def _get_instance_ip(subnet_hint: Optional[BaseIPNetwork] = None) -> str:
                if subnet_hint is not None and subnet_hint.prefixlen > 0:
                    local_ipaddrs = [*fetch_local_ipaddrs(subnet_hint)]
                    if local_ipaddrs:
                        return str(local_ipaddrs[0])
                    raise RuntimeError(
                        f"Could not find my IP address bound to subnet {subnet_hint}"
                    )
                try:
                    myself = socket.gethostname()
                    resolver = aiodns.DNSResolver()
                    result = await resolver.gethostbyname(myself, socket.AF_INET)
                    return result.addresses[0]
                except aiodns.error.DNSError:
                    return "127.0.0.1"

            async def _get_instance_type() -> str:
                return "default"

            async def _get_instance_region() -> str:
                return os.environ.get("BACKEND_REGION", "local")

    get_instance_id = _get_instance_id
    get_instance_ip = _get_instance_ip
    get_instance_type = _get_instance_type
    get_instance_region = _get_instance_region
    _defined = True


_define_functions()



from __future__ import annotations

import asyncio
import logging
import time
import uuid
from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from typing import (
    Awaitable,
    Callable,
    Concatenate,
    Final,
    Optional,
    Protocol,
    Self,
    TypeAlias,
)

from ai.backend.common.bgtask.types import (
    BackgroundTaskMetadata,
    BgtaskStatus,
    TaskID,
    TaskName,
)
from ai.backend.common.clients.valkey_client.valkey_bgtask.client import ValkeyBgtaskClient
from ai.backend.common.events.types import EventCacheDomain
from ai.backend.common.exception import (
    BackendAIError,
    BgtaskNotRegisteredError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)
from ai.backend.logging import BraceStyleAdapter

from ..events.dispatcher import (
    EventProducer,
)
from ..events.event_types.bgtask.broadcast import (
    BaseBgtaskDoneEvent,
    BgtaskCancelledEvent,
    BgtaskDoneEvent,
    BgtaskFailedEvent,
    BgtaskPartialSuccessEvent,
    BgtaskUpdatedEvent,
)
from ..types import DispatchResult, Sentinel
from .reporter import ProgressReporter
from .task.base import BaseBackgroundTaskArgs, BaseBackgroundTaskHandler
from .task.registry import BackgroundTaskHandlerRegistry

sentinel: Final = Sentinel.TOKEN
log = BraceStyleAdapter(logging.getLogger(__spec__.name))


BgtaskEvents: TypeAlias = (
    BgtaskUpdatedEvent
    | BgtaskDoneEvent
    | BgtaskCancelledEvent
    | BgtaskFailedEvent
    | BgtaskPartialSuccessEvent
)


_HEARTBEAT_INTERVAL = 60  # 1 minute
_HEARTBEAT_CHECK_INTERVAL = 300  # 5 minutes


@dataclass
class BgTaskInfo:
    status: BgtaskStatus
    msg: str
    started_at: str
    last_update: str
    current: str = "0"
    total: str = "0"

    @classmethod
    def started(cls, msg: str = "") -> Self:
        now = str(time.time())
        return cls(
            status=BgtaskStatus.STARTED,
            msg=msg,
            started_at=now,
            last_update=now,
            current="0",
            total="0",
        )

    @classmethod
    def finished(cls, status: BgtaskStatus, msg: str = "") -> Self:
        now = str(time.time())
        return cls(
            status=status,
            msg=msg,
            started_at="0",
            last_update=now,
            current="0",
            total="0",
        )

    def to_dict(self) -> Mapping[str | bytes, str]:
        return {
            "status": str(self.status),
            "msg": self.msg,
            "started_at": self.started_at,
            "last_update": self.last_update,
            "current": self.current,
            "total": self.total,
        }


BackgroundTask = Callable[
    Concatenate[ProgressReporter, ...], Awaitable[str | DispatchResult | None]
]


class BackgroundTaskObserver(Protocol):
    def observe_bgtask_started(self, *, task_name: str) -> None: ...
    def observe_bgtask_done(
        self, *, task_name: str, status: str, duration: float, error_code: Optional[ErrorCode]
    ) -> None: ...


class NopBackgroundTaskObserver:
    def observe_bgtask_started(self, *, task_name: str) -> None:
        pass

    def observe_bgtask_done(
        self, *, task_name: str, status: str, duration: float, error_code: Optional[ErrorCode]
    ) -> None:
        pass


class BackgroundTaskManager:
    _event_producer: EventProducer
    _ongoing_tasks: dict[TaskID, asyncio.Task]
    _metric_observer: BackgroundTaskObserver
    _dict_lock: asyncio.Lock

    _valkey_client: ValkeyBgtaskClient
    _server_id: str
    _tags: set[str]
    _task_registry: BackgroundTaskHandlerRegistry

    def __init__(
        self,
        event_producer: EventProducer,
        *,
        task_registry: Optional[BackgroundTaskHandlerRegistry] = None,
        valkey_client: ValkeyBgtaskClient,
        server_id: str,
        tags: Optional[Iterable[str]] = None,
        bgtask_observer: BackgroundTaskObserver = NopBackgroundTaskObserver(),
    ) -> None:
        self._event_producer = event_producer
        self._ongoing_tasks = {}
        self._metric_observer = bgtask_observer
        self._dict_lock = asyncio.Lock()

        self._valkey_client = valkey_client
        self._server_id = server_id
        self._tags = set(tags) if tags is not None else set()
        self._task_registry = task_registry or BackgroundTaskHandlerRegistry()

        self._heartbeat_loop_task = asyncio.create_task(self._heartbeat_loop())
        self._retry_loop_task = asyncio.create_task(self._retry_loop())

    async def start(
        self,
        func: BackgroundTask,
        name: Optional[str] = None,
        **kwargs,
    ) -> uuid.UUID:
        task_id = uuid.uuid4()
        await self._event_producer.broadcast_event_with_cache(
            EventCacheDomain.BGTASK.cache_id(str(task_id)),
            BgtaskUpdatedEvent(
                task_id=task_id,
                message="Task started",
                current_progress=0,
                total_progress=0,
            ),
        )
        task = asyncio.create_task(self._wrapper_task(func, task_id, name, **kwargs))
        self._ongoing_tasks[TaskID(task_id)] = task
        return task_id

    async def shutdown(self) -> None:
        log.info("Cancelling remaining background tasks...")
        for task in self._ongoing_tasks.values():
            if task.done():
                continue
            try:
                task.cancel()
                await task
            except asyncio.CancelledError:
                pass
        try:
            self._heartbeat_loop_task.cancel()
            await self._heartbeat_loop_task
        except asyncio.CancelledError:
            pass
        try:
            self._retry_loop_task.cancel()
            await self._retry_loop_task
        except asyncio.CancelledError:
            pass

    def _convert_bgtask_to_event(
        self, task_id: uuid.UUID, bgtask_result: DispatchResult | str | None
    ) -> BaseBgtaskDoneEvent:
        # legacy
        if bgtask_result is None or isinstance(bgtask_result, str):
            return BgtaskDoneEvent(task_id, bgtask_result)

        message = bgtask_result.message()
        if bgtask_result.has_error():
            return BgtaskPartialSuccessEvent(
                task_id=task_id, message=message, errors=bgtask_result.errors
            )
        else:
            return BgtaskDoneEvent(task_id=task_id, message=message)

    async def _run_bgtask(
        self,
        func: BackgroundTask,
        task_id: uuid.UUID,
        **kwargs,
    ) -> BaseBgtaskDoneEvent:
        reporter = ProgressReporter(self._event_producer, task_id)
        bgtask_result = await func(reporter, **kwargs)
        return self._convert_bgtask_to_event(task_id, bgtask_result)

    async def _observe_bgtask(
        self,
        func: BackgroundTask,
        task_id: uuid.UUID,
        task_name: Optional[str],
        **kwargs,
    ) -> BaseBgtaskDoneEvent:
        self._metric_observer.observe_bgtask_started(task_name=task_name or func.__name__)
        start_time = time.perf_counter()
        task_name = task_name or func.__name__
        status = BgtaskStatus.STARTED
        error_code: Optional[ErrorCode] = None
        msg = "no message"
        try:
            bgtask_result_event = await self._run_bgtask(func, task_id, **kwargs)
            status = bgtask_result_event.status()
            msg = bgtask_result_event.message or msg
        except asyncio.CancelledError:
            status = BgtaskStatus.CANCELLED
            error_code = ErrorCode(
                domain=ErrorDomain.BGTASK,
                operation=ErrorOperation.EXECUTE,
                error_detail=ErrorDetail.CANCELED,
            )
            log.warning("Task {} ({}): cancelled", task_id, task_name)
            msg = "Task cancelled"
            return BgtaskCancelledEvent(task_id, msg)
        except BackendAIError as e:
            status = BgtaskStatus.FAILED
            error_code = e.error_code()
            log.exception("Task {} ({}): BackendAIError: {}", task_id, task_name, e)
            msg = repr(e)
            return BgtaskFailedEvent(task_id, msg)
        except Exception as e:
            status = BgtaskStatus.FAILED
            error_code = ErrorCode(
                domain=ErrorDomain.BGTASK,
                operation=ErrorOperation.EXECUTE,
                error_detail=ErrorDetail.INTERNAL_ERROR,
            )
            log.exception("Task {} ({}): unhandled error: {}", task_id, task_name, e)
            msg = repr(e)
            return BgtaskFailedEvent(task_id, msg)
        finally:
            duration = time.perf_counter() - start_time
            self._metric_observer.observe_bgtask_done(
                task_name=task_name,
                status=status,
                duration=duration,
                error_code=error_code,
            )

        return bgtask_result_event

    async def _wrapper_task(
        self,
        func: BackgroundTask,
        task_id: uuid.UUID,
        task_name: Optional[str],
        **kwargs,
    ) -> None:
        try:
            bgtask_result_event = await self._observe_bgtask(func, task_id, task_name, **kwargs)
            cache_id = EventCacheDomain.BGTASK.cache_id(str(task_id))
            await self._event_producer.broadcast_event_with_cache(cache_id, bgtask_result_event)
            log.info(
                "Task {} ({}): {}", task_id, task_name or "", bgtask_result_event.__class__.__name__
            )
        finally:
            self._ongoing_tasks.pop(TaskID(task_id), None)

    def _get_task_func_by_name(self, name: TaskName) -> BaseBackgroundTaskHandler:
        return self._task_registry.get_task(name)

    async def start_retriable(
        self,
        task_name: TaskName,
        args: BaseBackgroundTaskArgs,
        tags: Optional[Iterable[str]] = None,
    ) -> TaskID:
        task_id = TaskID(uuid.uuid4())
        await self._event_producer.broadcast_event_with_cache(
            EventCacheDomain.BGTASK.cache_id(str(task_id)),
            BgtaskUpdatedEvent(
                task_id=task_id,
                message="Task started",
                current_progress=0,
                total_progress=0,
            ),
        )
        func = self._get_task_func_by_name(task_name)
        metadata = BackgroundTaskMetadata.create(
            task_id=task_id,
            task_name=task_name,
            body=args.to_metadata_body(),
            server_id=self._server_id,
            tags=tags,
        )
        task = asyncio.create_task(self._process_retriable_task(func, args, metadata))
        self._ongoing_tasks[task_id] = task
        return task_id

    async def _process_retriable_task(
        self,
        func: BaseBackgroundTaskHandler,
        args: BaseBackgroundTaskArgs,
        metadata: BackgroundTaskMetadata,
    ) -> None:
        await self._valkey_client.register_task(metadata)
        try:
            await self._wrapper_broadcast_result(
                func,
                args,
                metadata,
            )
        finally:
            self._ongoing_tasks.pop(metadata.task_id, None)
            await self._valkey_client.unregister_task(metadata)

    async def _wrapper_broadcast_result(
        self,
        func: BaseBackgroundTaskHandler,
        args: BaseBackgroundTaskArgs,
        metadata: BackgroundTaskMetadata,
    ) -> None:
        bgtask_result_event = await self._observe_retriable_bgtask(
            func,
            args,
            metadata,
        )
        cache_id = EventCacheDomain.BGTASK.cache_id(str(metadata.task_id))
        await self._event_producer.broadcast_event_with_cache(cache_id, bgtask_result_event)
        log.info(
            "Task {} ({}): {}",
            metadata.task_id,
            metadata.task_name or "",
            bgtask_result_event.__class__.__name__,
        )

    async def _observe_retriable_bgtask(
        self,
        func: BaseBackgroundTaskHandler,
        args: BaseBackgroundTaskArgs,
        metadata: BackgroundTaskMetadata,
    ) -> BaseBgtaskDoneEvent:
        task_name = func.name()
        task_id = metadata.task_id
        self._metric_observer.observe_bgtask_started(task_name=task_name)
        start_time = time.perf_counter()
        status = BgtaskStatus.STARTED
        error_code: Optional[ErrorCode] = None
        msg = "no message"
        try:
            result = await self._run_retriable_bgtask(func, args, metadata)
            bgtask_result_event = self._convert_result_to_event(result, metadata)
            status = bgtask_result_event.status()
            msg = bgtask_result_event.message or msg
        except asyncio.CancelledError:
            status = BgtaskStatus.CANCELLED
            error_code = ErrorCode(
                domain=ErrorDomain.BGTASK,
                operation=ErrorOperation.EXECUTE,
                error_detail=ErrorDetail.CANCELED,
            )
            log.warning("Task {} ({}): cancelled", task_id, task_name)
            msg = "Task cancelled"
            return BgtaskCancelledEvent(task_id, msg)
        except BackendAIError as e:
            status = BgtaskStatus.FAILED
            error_code = e.error_code()
            log.exception("Task {} ({}): BackendAIError: {}", task_id, task_name, e)
            msg = repr(e)
            return BgtaskFailedEvent(task_id, msg)
        except Exception as e:
            status = BgtaskStatus.FAILED
            error_code = ErrorCode(
                domain=ErrorDomain.BGTASK,
                operation=ErrorOperation.EXECUTE,
                error_detail=ErrorDetail.INTERNAL_ERROR,
            )
            log.exception("Task {} ({}): unhandled error: {}", task_id, task_name, e)
            msg = repr(e)
            return BgtaskFailedEvent(task_id, msg)
        finally:
            duration = time.perf_counter() - start_time
            self._metric_observer.observe_bgtask_done(
                task_name=task_name,
                status=status,
                duration=duration,
                error_code=error_code,
            )

        return bgtask_result_event

    async def _run_retriable_bgtask(
        self,
        func: BaseBackgroundTaskHandler,
        args: BaseBackgroundTaskArgs,
        metadata: BackgroundTaskMetadata,
    ) -> DispatchResult:
        reporter = ProgressReporter(self._event_producer, metadata.task_id)
        return await func.execute(reporter, args)

    def _convert_result_to_event(
        self,
        bgtask_result: DispatchResult,
        metadata: BackgroundTaskMetadata,
    ) -> BaseBgtaskDoneEvent:
        task_id = metadata.task_id
        message = bgtask_result.message()
        if bgtask_result.has_error():
            return BgtaskPartialSuccessEvent(
                task_id=task_id, message=message, errors=bgtask_result.errors
            )
        else:
            return BgtaskDoneEvent(task_id=task_id, message=message)

    async def _heartbeat_loop(self) -> None:
        """Periodically update heartbeat for running background tasks"""
        while True:
            try:
                # Update heartbeat for all ongoing background tasks
                alive_task_ids: list[TaskID] = []
                for task_id, bg_task in self._ongoing_tasks.items():
                    if not bg_task.done():
                        alive_task_ids.append(task_id)
                await self._valkey_client.heartbeat(
                    alive_task_ids,
                    server_id=self._server_id,
                    tags=self._tags,
                )
            except Exception as e:
                log.exception("Exception in heartbeat loop: {}", e)
            await asyncio.sleep(_HEARTBEAT_INTERVAL)

    async def _retry_loop(self) -> None:
        """Main recovery loop that checks for failed/stale tasks"""
        while True:
            try:
                await self._check_server_tasks()
                await self._check_tagged_tasks()
            except Exception as e:
                log.exception("Exception in retry loop: {}", e)
            await asyncio.sleep(_HEARTBEAT_CHECK_INTERVAL)

    async def _check_server_tasks(self) -> None:
        timeout_task_metadata = await self._valkey_client.list_timeout_tasks_by_server_id(
            self._server_id
        )

        for metadata in timeout_task_metadata:
            await self._retry_bgtask(metadata)

    async def _check_tagged_tasks(self) -> None:
        """Check tasks for a specific server type"""
        timeout_task_metadata = await self._valkey_client.list_timeout_tasks_by_tags(self._tags)

        for metadata in timeout_task_metadata:
            await self._retry_bgtask(metadata)

    async def _retry_bgtask(self, metadata: BackgroundTaskMetadata) -> None:
        """Retry a background task"""

        metadata.server_id = self._server_id  # Claim the task
        task_name = metadata.task_name
        try:
            func = self._get_task_func_by_name(task_name)
        except BgtaskNotRegisteredError:
            log.exception(
                "Task {} ({}): not registered in {} server, skipping retry",
                metadata.task_id,
                task_name,
                self._server_id,
            )
            return
        args_type = func.args_type()
        args = args_type.from_metadata_body(metadata.body)
        retry_task = self._process_retriable_task(func, args, metadata=metadata)
        task = asyncio.create_task(retry_task)
        self._ongoing_tasks[metadata.task_id] = task



import enum
from typing import Optional, Self


class KernelLifecycleEventReason(enum.StrEnum):
    AGENT_TERMINATION = "agent-termination"
    ALREADY_TERMINATED = "already-terminated"
    ANOMALY_DETECTED = "anomaly-detected"
    EXEC_TIMEOUT = "exec-timeout"
    FAILED_TO_CREATE = "failed-to-create"
    FAILED_TO_START = "failed-to-start"
    FORCE_TERMINATED = "force-terminated"
    BOOTSTRAP_TIMEOUT = "bootstrap-timeout"
    HANG_TIMEOUT = "hang-timeout"
    IDLE_TIMEOUT = "idle-timeout"
    IDLE_SESSION_LIFETIME = "idle-session-lifetime"
    IDLE_UTILIZATION = "idle-utilization"
    KILLED_BY_EVENT = "killed-by-event"
    SERVICE_SCALED_DOWN = "service-scaled-down"
    NEW_CONTAINER_STARTED = "new-container-started"
    PENDING_TIMEOUT = "pending-timeout"
    RESTARTING = "restarting"
    RESTART_TIMEOUT = "restart-timeout"
    RESUMING_AGENT_OPERATION = "resuming-agent-operation"
    SELF_TERMINATED = "self-terminated"
    TASK_FAILED = "task-failed"
    TASK_TIMEOUT = "task-timeout"
    TASK_CANCELLED = "task-cancelled"
    TASK_FINISHED = "task-finished"
    TERMINATED_UNKNOWN_CONTAINER = "terminated-unknown-container"
    UNKNOWN = "unknown"
    USER_REQUESTED = "user-requested"
    USER_PURGED = "user-purged"
    NOT_FOUND_IN_MANAGER = "not-found-in-manager"
    CONTAINER_NOT_FOUND = "container-not-found"

    @classmethod
    def from_value(cls, value: Optional[str]) -> Optional[Self]:
        if value is None:
            return None
        try:
            return cls(value)
        except ValueError:
            pass
        return None



"""Base classes for periodic tasks."""

from __future__ import annotations

from abc import ABC, abstractmethod


class PeriodicTask(ABC):
    """Abstract base class for tasks that execute periodically when leader."""

    @abstractmethod
    async def run(self) -> None:
        """Execute the task."""
        raise NotImplementedError

    @property
    @abstractmethod
    def name(self) -> str:
        """Task name for logging and identification."""
        raise NotImplementedError

    @property
    @abstractmethod
    def interval(self) -> float:
        """Interval between task executions in seconds."""
        raise NotImplementedError

    @property
    @abstractmethod
    def initial_delay(self) -> float:
        """Initial delay before first execution in seconds."""
        raise NotImplementedError



from __future__ import annotations

import enum
import logging
import textwrap
import uuid
from collections.abc import Collection
from typing import Optional, Self, cast

from glide import (
    Batch,
    ExpirySet,
    ExpiryType,
    Script,
)

from ai.backend.common.bgtask.exception import InvalidTaskMetadataError
from ai.backend.common.bgtask.types import (
    BackgroundTaskMetadata,
    TaskID,
)
from ai.backend.common.clients.valkey_client.client import (
    AbstractValkeyClient,
    create_layer_aware_valkey_decorator,
    create_valkey_client,
)
from ai.backend.common.data.bgtask.defs import (
    TASK_METADATA_TTL,
    TASK_TTL_THRESHOLD,
)
from ai.backend.common.defs import REDIS_BGTASK_DB
from ai.backend.common.metrics.metric import LayerType
from ai.backend.common.types import ValkeyTarget
from ai.backend.logging.utils import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

# Layer-specific decorator for valkey_bgtask client
valkey_decorator = create_layer_aware_valkey_decorator(LayerType.VALKEY_BGTASK)

_KEY_PREFIX = "bgtask"
_TASK_KEY_PREFIX = f"{_KEY_PREFIX}:task"  # bgtask:task:{task_id}
_TAG_KEY_PREFIX = f"{_KEY_PREFIX}:tag"  # bgtask:tag:{tag}
_SERVER_KEY_PREFIX = f"{_KEY_PREFIX}:server"  # bgtask:server:{server_id}


class _ScriptResult(enum.StrEnum):
    KEY_NOT_EXIST = "key_not_exist"
    NO_EXPIRY = "no_expiry"
    TTL_SUFFICIENT = "ttl_sufficient"
    TTL_INSUFFICIENT = "ttl_insufficient"

    @classmethod
    def from_bytes(cls, value: bytes) -> _ScriptResult:
        return cls(value.decode())


class ValkeyBgtaskClient:
    """
    Client for background task management using Valkey/Glide.
    """

    _client: AbstractValkeyClient
    _closed: bool

    def __init__(
        self,
        client: AbstractValkeyClient,
    ) -> None:
        self._client = client
        self._closed = False

    @classmethod
    async def create(
        cls,
        valkey_target: ValkeyTarget,
        *,
        db_id: int = REDIS_BGTASK_DB,
        human_readable_name: str = "bgtask",
    ) -> Self:
        """
        Create a ValkeyBgtaskClient instance.
        """
        client = create_valkey_client(
            valkey_target,
            db_id=db_id,
            human_readable_name=human_readable_name,
        )
        await client.connect()
        return cls(client)

    async def close(self) -> None:
        """Close the client connection."""
        if not self._closed:
            await self._client.disconnect()
            self._closed = True

    def _get_tag_key(self, tag: str) -> str:
        return f"{_TAG_KEY_PREFIX}:{tag}"

    def _get_server_key(self, server_id: str) -> str:
        return f"{_SERVER_KEY_PREFIX}:{server_id}"

    def _get_task_key(self, task_id: TaskID) -> str:
        return f"{_TASK_KEY_PREFIX}:{task_id}"

    # Task metadata operations
    @valkey_decorator()
    async def register_task(self, metadata: BackgroundTaskMetadata) -> None:
        """
        Register a background task with 24-hour TTL and index it by tags and server ID.
        """
        batch = self._create_batch()
        key = self._get_task_key(metadata.task_id)
        value = metadata.to_json()
        batch.set(key, value, expiry=ExpirySet(ExpiryType.SEC, TASK_METADATA_TTL))

        for tag in metadata.tags:
            tag_key = self._get_tag_key(tag)
            batch.sadd(tag_key, [metadata.task_id.hex])
            batch.expire(tag_key, TASK_METADATA_TTL)

        server_key = self._get_server_key(metadata.server_id)
        batch.sadd(server_key, [metadata.task_id.hex])
        batch.expire(server_key, TASK_METADATA_TTL)
        await self._client.client.exec(batch, raise_on_error=True)

    @valkey_decorator()
    async def unregister_task(self, metadata: BackgroundTaskMetadata) -> None:
        """
        Remove task and all its index references. Idempotent operation.
        """
        batch = self._create_batch()
        key = self._get_task_key(metadata.task_id)
        batch.delete([key])

        for tag in metadata.tags:
            tag_key = self._get_tag_key(tag)
            batch.srem(tag_key, [metadata.task_id.hex])

        server_key = self._get_server_key(metadata.server_id)
        batch.srem(server_key, [metadata.task_id.hex])
        await self._client.client.exec(batch, raise_on_error=True)

    async def _get_registered_task_ids(self, keys: Collection[str]) -> set[TaskID]:
        if not keys:
            return set()
        batch = self._create_batch()
        for key in keys:
            batch.smembers(key)
        raw_results = await self._client.client.exec(batch, raise_on_error=True)
        if raw_results is None:
            raise RuntimeError("Failed to retrieve members from keys")
        results = cast(list[Optional[set[bytes]]], raw_results)
        task_ids: set[TaskID] = set()
        for raw_result in results:
            if raw_result is not None:
                task_ids |= {TaskID(uuid.UUID(hex=hex_id.decode())) for hex_id in raw_result}
        return task_ids

    @valkey_decorator()
    async def list_timeout_tasks_by_tags(
        self, tags: Collection[str]
    ) -> list[BackgroundTaskMetadata]:
        """
        List tasks with insufficient TTL filtered by tags.
        """
        if not tags:
            return []
        keys = [self._get_tag_key(tag) for tag in tags]
        task_ids = await self._get_registered_task_ids(keys)
        results = await self._list_timeout_tasks(task_ids)
        return results

    @valkey_decorator()
    async def list_timeout_tasks_by_server_id(self, server_id: str) -> list[BackgroundTaskMetadata]:
        """
        List tasks with insufficient TTL owned by a specific server.
        """
        key = self._get_server_key(server_id)
        task_ids = await self._get_registered_task_ids([key])
        results = await self._list_timeout_tasks(task_ids)
        return results

    async def _list_timeout_tasks(
        self, task_ids: Collection[TaskID]
    ) -> list[BackgroundTaskMetadata]:
        if not task_ids:
            return []
        script = self._task_getter_script()
        results: list[BackgroundTaskMetadata] = []
        for task_id in task_ids:
            key = self._get_task_key(task_id)
            raw_result = await self._client.client.invoke_script(
                script,
                keys=[key],
                args=[
                    str(TASK_TTL_THRESHOLD),
                    str(TASK_METADATA_TTL),
                ],
            )
            result = cast(list[bytes], raw_result)
            result_type, metadata = self._resolve_script_result(result)
            match result_type:
                case _ScriptResult.TTL_SUFFICIENT:
                    log.debug(
                        "Task TTL sufficient, skipping (id: {})",
                        task_id,
                    )
                    continue
                case _ScriptResult.KEY_NOT_EXIST | _ScriptResult.NO_EXPIRY:
                    log.warning(
                        "Task key not exist or no expiry, skipping (id: {}, result: {})",
                        task_id,
                        result_type,
                    )
                    continue
                case _ScriptResult.TTL_INSUFFICIENT:
                    log.info(
                        "Task TTL insufficient, fetching metadata (id: {}, metadata: {})",
                        task_id,
                        metadata,
                    )
                    if metadata is None:
                        continue

            results.append(metadata)

        return results

    def _resolve_script_result(
        self, raw_result: list[bytes]
    ) -> tuple[_ScriptResult, Optional[BackgroundTaskMetadata]]:
        result_type = _ScriptResult.from_bytes(raw_result[0])
        metadata: Optional[BackgroundTaskMetadata] = None
        if len(raw_result) == 2:
            raw_metadata = raw_result[1]
            try:
                metadata = BackgroundTaskMetadata.from_json(raw_metadata)
            except InvalidTaskMetadataError:
                log.exception("Invalid bgtask metadata (data: {})", raw_metadata)
        return result_type, metadata

    @valkey_decorator()
    async def heartbeat(
        self,
        task_ids: Collection[TaskID],
        server_id: Optional[str],
        tags: Collection[str],
    ) -> None:
        """
        Extend TTL to 24 hours for active tasks. Non-existent tasks are ignored.
        """
        keys: list[str] = []

        task_keys = [self._get_task_key(task_id) for task_id in task_ids]
        keys.extend(task_keys)
        if server_id is not None:
            server_key = self._get_server_key(server_id)
            keys.append(server_key)
        tag_keys = [self._get_tag_key(tag) for tag in tags]
        keys.extend(tag_keys)

        if not keys:
            return
        batch = self._create_batch()
        for key in keys:
            batch.expire(key, TASK_METADATA_TTL)
        await self._client.client.exec(batch, raise_on_error=True)

    def _create_batch(self, is_atomic: bool = False) -> Batch:
        return Batch(is_atomic=is_atomic)

    def _task_getter_script(self) -> Script:
        code = textwrap.dedent(f"""
        -- KEYS[1]: Key
        -- ARGV[1]: TTL threshold (in seconds)
        -- ARGV[2]: New TTL value to set (in seconds)

        local target_key = KEYS[1]
        local ttl_threshold = tonumber(ARGV[1])
        local new_ttl = tonumber(ARGV[2])

        -- Check if key exists and get current TTL
        local current_ttl = redis.call('TTL', target_key)

        -- Handle non-existent key (-2)
        if current_ttl == -2 then
            -- Key doesn't exist
            return {{'{_ScriptResult.KEY_NOT_EXIST}'}}
        end

        -- Key exists, set new TTL
        redis.call('EXPIRE', target_key, new_ttl)

        -- Handle key without expiration (-1) - treat as infinite TTL
        if current_ttl == -1 then
            -- Key exists but has no expiration, TTL is sufficient
            return {{'{_ScriptResult.NO_EXPIRY}'}}
        end

        -- Key exists and has TTL, check threshold
        if current_ttl < ttl_threshold then
            -- TTL is below threshold, return true with value
            local value = redis.call('GET', target_key)
            return {{'{_ScriptResult.TTL_INSUFFICIENT}', value}}
        else
            -- TTL is sufficient, return false with nil
            return {{'{_ScriptResult.TTL_SUFFICIENT}'}}
        end
        """)
        return Script(code)



from __future__ import annotations

import enum
from typing import Any, Self

from pydantic import BaseModel


class AuthTokenType(enum.StrEnum):
    KEYPAIR = "keypair"
    JWT = "jwt"


class AuthResponseType(enum.StrEnum):
    SUCCESS = "success"
    REQUIRE_TWO_FACTOR_REGISTRATION = "REQUIRE_TWO_FACTOR_REGISTRATION"
    REQUIRE_TWO_FACTOR_AUTH = "REQUIRE_TWO_FACTOR_AUTH"


class TwoFactorType(enum.StrEnum):
    TOTP = "TOTP"


class AuthResponse(BaseModel):
    response_type: AuthResponseType

    @classmethod
    def parse(cls, data: dict[str, Any]) -> Self:
        return cls.model_validate(data)


class AuthSuccessResponse(AuthResponse):
    access_key: str
    secret_key: str
    role: str
    status: str
    type: AuthTokenType = AuthTokenType.KEYPAIR

    def to_dict(self) -> dict[str, str]:
        return self.model_dump(mode="json")


class RequireTwoFactorRegistrationResponse(AuthResponse):
    token: str
    type: TwoFactorType

    def to_dict(self) -> dict[str, str]:
        return self.model_dump(mode="json")


class RequireTwoFactorAuthResponse(AuthResponse):
    type: TwoFactorType

    def to_dict(self) -> dict[str, str]:
        return self.model_dump(mode="json")


def parse_auth_response(data: dict[str, Any]) -> AuthResponse:
    raw_response_type = data.get("response_type")
    respones_type = (
        AuthResponseType(raw_response_type)
        if raw_response_type is not None
        else AuthResponseType.SUCCESS
    )
    match respones_type:
        case AuthResponseType.SUCCESS:
            return AuthSuccessResponse.model_validate(data)
        case AuthResponseType.REQUIRE_TWO_FACTOR_REGISTRATION:
            return RequireTwoFactorRegistrationResponse.model_validate(data)
        case AuthResponseType.REQUIRE_TWO_FACTOR_AUTH:
            return RequireTwoFactorAuthResponse.model_validate(data)



from enum import StrEnum


class EndpointStatus(StrEnum):
    READY = "READY"
    PROVISIONING = "PROVISIONING"
    HEALTHY = "HEALTHY"
    UNHEALTHY = "UNHEALTHY"
    DESTROYING = "DESTROYING"
    DESTROYED = "DESTROYED"
    DEGRADED = "DEGRADED"



from typing import Optional

from pydantic import AliasChoices, BaseModel, Field, field_serializer, field_validator

from ai.backend.common.typed_validators import HostPortPair as HostPortPairModel
from ai.backend.common.types import RedisHelperConfig as RedisHelperConfigDict
from ai.backend.common.types import RedisProfileTarget, ValkeyProfileTarget, ValkeyTarget
from ai.backend.common.types import RedisTarget as _RedisTarget


class RedisHelperConfig(BaseModel):
    socket_timeout: float = Field(
        default=5.0,
        description="""
        Timeout in seconds for Redis socket operations.
        Controls how long operations wait before timing out.
        Increase for slow or congested networks.
        """,
        examples=[5.0, 10.0],
        validation_alias=AliasChoices("socket_timeout", "socket-timeout"),
        serialization_alias="socket_timeout",
    )
    socket_connect_timeout: float = Field(
        default=2.0,
        description="""
        Timeout in seconds for establishing Redis connections.
        Controls how long connection attempts wait before failing.
        Shorter values fail faster but may be too aggressive for some networks.
        """,
        examples=[2.0, 5.0],
        validation_alias=AliasChoices("socket_connect_timeout", "socket-connect-timeout"),
        serialization_alias="socket_connect_timeout",
    )
    reconnect_poll_timeout: float = Field(
        default=0.3,
        description="""
        Time in seconds to wait between reconnection attempts.
        Controls the polling frequency when trying to reconnect to Redis.
        Lower values reconnect faster but may increase network load.
        """,
        examples=[0.3, 1.0],
        validation_alias=AliasChoices("reconnect_poll_timeout", "reconnect-poll-timeout"),
        serialization_alias="reconnect_poll_timeout",
    )

    def to_dict(self) -> RedisHelperConfigDict:
        return RedisHelperConfigDict(
            socket_timeout=self.socket_timeout,
            socket_connect_timeout=self.socket_connect_timeout,
            reconnect_poll_timeout=self.reconnect_poll_timeout,
        )


class SingleRedisConfig(BaseModel):
    addr: Optional[HostPortPairModel] = Field(
        default=None,
        description="""
        Network address and port of the Redis server.
        Redis is used for distributed caching and messaging between managers.
        Set to None when using Sentinel for high availability.
        """,
        examples=[None, {"host": "127.0.0.1", "port": 6379}],
    )
    sentinel: Optional[list[HostPortPairModel]] = Field(
        default=None,
        description="""
        List of Redis Sentinel addresses for high availability.
        If provided, the manager will use Redis Sentinel for automatic failover.
        When using Sentinel, the addr field is ignored and service_name is required.
        """,
        examples=[
            None,
            [{"host": "redis-sentinel", "port": 26379}, {"host": "redis-sentinel", "port": 26380}],
        ],
    )
    service_name: Optional[str] = Field(
        default=None,
        description="""
        Service name for Redis Sentinel.
        Required when using Redis Sentinel for high availability.
        Identifies which service to monitor for failover.
        """,
        examples=[None, "mymaster", "backend-ai"],
        validation_alias=AliasChoices("service_name", "service-name"),
        serialization_alias="service-name",
    )
    password: Optional[str] = Field(
        default=None,
        description="""
        Password for authenticating with Redis.
        Set to None if Redis doesn't require authentication.
        Should be kept secret in production environments.
        """,
        examples=[None, "REDIS_PASSWORD"],
    )
    request_timeout: Optional[int] = Field(
        default=None,
        description="""
        Timeout in milliseconds for Redis requests.
        Controls how long operations wait before timing out.
        If None, uses the default timeout configured in the Redis client.
        """,
        examples=[None, 1000],
        validation_alias=AliasChoices("request_timeout", "request-timeout"),
        serialization_alias="request-timeout",
    )
    redis_helper_config: RedisHelperConfig = Field(
        default_factory=RedisHelperConfig,
        description="""
        Configuration for the Redis helper library.
        Controls timeouts and reconnection behavior.
        Adjust based on network conditions and reliability requirements.
        """,
        validation_alias=AliasChoices("redis_helper_config", "redis-helper-config"),
        serialization_alias="redis-helper-config",
    )
    use_tls: bool = Field(
        default=False,
        description="""
        Whether to use TLS for Redis connections.""",
        validation_alias=AliasChoices("use_tls", "use-tls"),
    )
    tls_skip_verify: bool = Field(
        default=False,
        description="""
        Whether to skip TLS certificate verification.
        Set to True for self-signed certificates or development environments.
        """,
        validation_alias=AliasChoices("tls_skip_verify", "tls-skip-verify"),
    )

    @field_validator("sentinel", mode="before")
    @classmethod
    def _parse_sentinel(
        cls, v: Optional[str | list[HostPortPairModel]]
    ) -> Optional[list[HostPortPairModel]]:
        if v is None or isinstance(v, list):
            return v
        if isinstance(v, str):
            entries: list[HostPortPairModel] = []
            for part in v.split(","):
                host, port = part.strip().split(":")
                entries.append(HostPortPairModel(host=host, port=int(port)))
            return entries
        raise TypeError("sentinel must be list or 'host:port,host:port' string")

    @field_serializer("addr")
    def _serialize_addr(self, addr: Optional[HostPortPairModel], _info) -> Optional[str]:
        return None if addr is None else f"{addr.host}:{addr.port}"

    @field_serializer("sentinel")
    def _serialize_sentinel(
        self, sentinel: Optional[list[HostPortPairModel]], _info
    ) -> Optional[str]:
        if sentinel is None:
            return None
        return ",".join(f"{hp.host}:{hp.port}" for hp in sentinel)

    def to_redis_target(self) -> _RedisTarget:
        """
        Convert this Redis configuration to a Redis target.
        This is used for serialization in the Backend.AI profile system.
        """
        addr = self.addr.to_legacy() if self.addr else None
        sentinel_addrs = None
        if self.sentinel:
            sentinel_addrs = [hp.to_legacy() for hp in self.sentinel]

        return _RedisTarget(
            addr=addr,
            sentinel=sentinel_addrs,
            service_name=self.service_name,
            password=self.password,
            redis_helper_config=self.redis_helper_config.to_dict(),
        )

    def to_valkey_target(self) -> ValkeyTarget:
        """
        Convert this Redis configuration to a Valkey target.
        This is used for serialization in the Backend.AI profile system.
        """
        addr = self.addr.address if self.addr else None
        sentinel_addrs = None
        if self.sentinel:
            sentinel_addrs = [hp.address for hp in self.sentinel]

        return ValkeyTarget(
            addr=addr,
            sentinel=sentinel_addrs,
            service_name=self.service_name,
            password=self.password,
            request_timeout=self.request_timeout,
        )


class RedisConfig(SingleRedisConfig):
    override_configs: Optional[dict[str, SingleRedisConfig]] = Field(
        default=None,
        description="""
        Optional override configurations for specific Redis contexts.
        Allows different Redis settings for different services within Backend.AI.
        Each key represents a context name, and the value is a complete Redis configuration.
        """,
        examples=[
            None,
            {
                "live": {
                    "addr": {"host": "127.0.0.1", "port": 6379},
                    "password": "different-password",
                    "redis_helper_config": {"socket_timeout": 10.0},
                }
            },
        ],
        validation_alias=AliasChoices("override_configs", "override-configs"),
        serialization_alias="override-configs",
    )

    def to_redis_profile_target(self) -> RedisProfileTarget:
        """
        Convert this Redis configuration to a profile target dictionary.
        This is used for serialization in the Backend.AI profile system.
        """
        addr = self.addr.to_legacy() if self.addr else None
        sentinel_addrs = None
        if self.sentinel:
            sentinel_addrs = [hp.to_legacy() for hp in self.sentinel]

        override_targets: Optional[dict[str, _RedisTarget]] = None
        if self.override_configs:
            override_targets = {k: v.to_redis_target() for k, v in self.override_configs.items()}

        return RedisProfileTarget(
            addr=addr,
            sentinel=sentinel_addrs,
            service_name=self.service_name,
            password=self.password,
            redis_helper_config=self.redis_helper_config.to_dict(),
            override_targets=override_targets,
        )

    def to_valkey_profile_target(self) -> ValkeyProfileTarget:
        """
        Convert this Valkey configuration to a profile target dictionary.
        This is used for serialization in the Backend.AI profile system.
        """
        addr = self.addr.address if self.addr else None
        sentinel_addrs = None
        if self.sentinel:
            sentinel_addrs = [hp.address for hp in self.sentinel]
        override_targets: Optional[dict[str, ValkeyTarget]] = None
        if self.override_configs:
            override_targets = {k: v.to_valkey_target() for k, v in self.override_configs.items()}

        return ValkeyProfileTarget(
            addr=addr,
            sentinel=sentinel_addrs,
            service_name=self.service_name,
            password=self.password,
            override_targets=override_targets,
            request_timeout=self.request_timeout,
        )



from typing import override

from ..exception import BackendAIError, ErrorCode, ErrorDetail, ErrorDomain, ErrorOperation


class InvalidTaskMetadataError(BackendAIError):
    error_type = "https://api.backend.ai/probs/invalid-bgtask-metadata"
    error_title = "Invalid Task Metadata"

    @classmethod
    @override
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BGTASK,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.UNREACHABLE,
        )



from pydantic import BaseModel, Field


class BgTaskProgressData(BaseModel):
    current: int = Field(
        default=0,
        description="Current progress of the scan operation, expressed as a percentage.",
        examples=[0, 50, 100],
    )
    total: int = Field(
        default=0,
        description="Total number of items to be scanned, used to calculate progress.",
        examples=[100, 200, 0],
    )



import uuid
from dataclasses import dataclass
from typing import Any, Sequence, cast

from ai.backend.common.etcd import AsyncEtcd, ConfigScopes

from ..service_discovery import ServiceDiscovery, ServiceMetadata

_DEFAULT_PREFIX = "ai/backend/service_discovery"


@dataclass
class ETCDServiceDiscoveryArgs:
    etcd: AsyncEtcd
    prefix: str = _DEFAULT_PREFIX


class ETCDServiceDiscovery(ServiceDiscovery):
    _etcd: AsyncEtcd
    _prefix: str

    def __init__(self, args: ETCDServiceDiscoveryArgs) -> None:
        self._etcd = args.etcd
        self._prefix = args.prefix

    async def register(self, service_meta: ServiceMetadata) -> None:
        prefix = self._service_prefix(service_meta.service_group, service_meta.id)
        await self._etcd.put_prefix(prefix, service_meta.to_dict(), scope=ConfigScopes.GLOBAL)

    async def unregister(self, service_group: str, service_id: uuid.UUID) -> None:
        await self._etcd.delete_prefix(
            self._service_prefix(service_group, service_id), scope=ConfigScopes.GLOBAL
        )

    async def heartbeat(self, service_meta: ServiceMetadata) -> None:
        service_meta.health_status.update_heartbeat()
        prefix = self._service_prefix(service_meta.service_group, service_meta.id)
        await self._etcd.put_prefix(prefix, service_meta.to_dict(), scope=ConfigScopes.GLOBAL)

    async def discover(self) -> Sequence[ServiceMetadata]:
        raw_service_groups = await self._etcd.get_prefix(self._prefix, scope=ConfigScopes.GLOBAL)
        services = []
        if not raw_service_groups:
            raise ValueError("No service groups found.")
        service_groups = cast(dict[str, dict[str, dict[str, Any]]], raw_service_groups)
        for _, service_configs in service_groups.items():
            for _, service_config in service_configs.items():
                services.append(ServiceMetadata.from_dict(service_config))
        return services

    async def get_service_group(self, service_group: str) -> Sequence[ServiceMetadata]:
        service_group_prefix = self._service_group_prefix(service_group)
        raw_service_configs = await self._etcd.get_prefix(
            service_group_prefix, scope=ConfigScopes.GLOBAL
        )
        if not raw_service_configs:
            raise ValueError(f"No services found in group {service_group}.")
        service_configs = cast(dict[str, dict[str, Any]], raw_service_configs)
        services = []
        for _, service_config in service_configs.items():
            services.append(ServiceMetadata.from_dict(service_config))
        return services

    async def get_service(self, service_group: str, service_id: uuid.UUID) -> ServiceMetadata:
        service_prefix = self._service_prefix(service_group, service_id)
        raw_service_config = await self._etcd.get_prefix(service_prefix, scope=ConfigScopes.GLOBAL)
        if not raw_service_config:
            raise ValueError(f"Service with ID {service_id} not found.")
        service_config = cast(dict[str, Any], raw_service_config)
        return ServiceMetadata.from_dict(service_config)

    def _service_group_prefix(self, service_group: str) -> str:
        return f"{self._prefix}/{service_group}"

    def _service_prefix(self, service_group: str, service_id: uuid.UUID) -> str:
        return f"{self._service_group_prefix(service_group)}/{service_id}"



"""
Sample configuration generator for Backend.AI.

This module provides utilities to generate sample TOML configuration files
from Pydantic models by extracting field information, defaults, descriptions,
and examples using JSON Schema.
"""

import enum
import logging
import textwrap
from collections.abc import Sequence
from dataclasses import dataclass
from datetime import timedelta
from ipaddress import IPv4Address, IPv4Network, IPv6Address, IPv6Network
from pathlib import PosixPath, PurePath
from typing import Any, Optional, Type

import toml
from pydantic import BaseModel
from pydantic_core import PydanticUndefined
from toml.decoder import InlineTableDict
from toml.encoder import TomlPreserveInlineDictEncoder
from yarl import URL

from ai.backend.common.typed_validators import HostPortPair
from ai.backend.common.types import BinarySize
from ai.backend.logging import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__name__))

_base_values_by_type = {
    "string": "...",
    "integer": 0,
    "number": 0.0,
    "boolean": False,
    "array": [],
    "object": {},
}


@dataclass(slots=True)
class FormatterContext:
    hint: str = ""
    annotation: type | None = None


class _InlineTable(dict, InlineTableDict):
    pass


def _wrap_comment(text: str, prefix: str = "", width: int = 80) -> str:
    """Wrap text into multiline comment format."""
    lines = text.strip().split("\n")
    wrapped_lines = []

    for line in lines:
        line = line.strip()
        if not line:
            wrapped_lines.append("#")
        else:
            # Wrap long lines
            wrapped = textwrap.wrap(line, width=width - 2)  # Account for '# '
            for wrapped_line in wrapped:
                wrapped_lines.append(f"{prefix}# {wrapped_line}")

    return "\n".join(wrapped_lines)


def _dump_toml_scalar(
    value: Any,
    default: Any,
    ctx: FormatterContext | None = None,
) -> str:
    match value:
        case {"$ref": complex_type}:
            type_name = complex_type.removeprefix("#/$defs/").split("__")[-1]
            if default is not None:
                return _dump_toml_scalar(default, None, ctx)
            return _dump_toml_scalar("{ " + type_name + " }", default, ctx)
        case {"type": "array", "items": item_type}:
            if default is not None:
                if isinstance(default, list):
                    return _dump_toml_scalar(default, None, ctx)
                else:
                    return "[ " + _dump_toml_scalar(default, None, ctx) + " ]"
            return "[ " + _dump_toml_scalar(item_type, default, ctx) + " ]"
        case {"type": "integer"} | {"type": "number"}:
            if default is None:
                out = "  "
            else:
                out = _dump_toml_scalar(default, None, ctx)
            has_min = False
            if (min_ := value.get("minimum", None)) is not None:
                out += f"  # min={min_}"
                has_min = True
            if (max_ := value.get("maximum", None)) is not None:
                if not has_min:
                    out += "  #"
                out += f" max={max_}"
            return out
        case {"type": "string", "format": "file-path"} | {"type": "string", "format": "path"}:
            if default is not None:
                return _dump_toml_scalar(default, None, ctx)
            return _dump_toml_scalar("PATH", default)
        case {"type": toml_type}:
            if default is not None:
                return _dump_toml_scalar(default, None, ctx)
            return _dump_toml_scalar(_base_values_by_type[toml_type], default)
        case IPv4Network() | IPv4Address() | IPv6Network() | IPv6Address():
            if default is not None:
                return _dump_toml_scalar(default, None, ctx)
            return str(value)
        case PosixPath() | PurePath() | URL() | timedelta():
            if default is not None:
                return _dump_toml_scalar(default, None, ctx)
            return str(value)
    if default is not None:
        return _dump_toml_scalar(default, None, ctx)
    if ctx is not None:
        match ctx.hint:
            case "BinarySize":
                value = f"{BinarySize(value):s}".upper()
            case "HostPortPair":
                value = {"host": value.host, "port": value.port}
            case "EnumByValue":
                assert ctx.annotation is not None
                value = ctx.annotation(value).value
            case "EnumByName":
                assert ctx.annotation is not None
                value = ctx.annotation(value).name
    if isinstance(value, dict):
        return (
            toml.dumps({"x": _InlineTable(value)}, encoder=TomlPreserveInlineDictEncoder())
            .strip()
            .split(" = ", 1)[1]
        )
    return toml.dumps({"x": value}).strip().split(" = ", 1)[1]


def _generate_sample_config(model_class: Type[BaseModel]) -> str:
    """
    Generate a sample TOML configuration file from a Pydantic model.

    Args:
        model_class: The Pydantic model class to generate configuration for

    Returns:
        A string containing the TOML configuration with comments
    """
    warnings: list[str] = []
    schema = model_class.model_json_schema()

    def _get_field_info(model_cls: Type[BaseModel], field_name: str, indent: int) -> dict[str, Any]:
        """Extract field information from Pydantic model including Field metadata."""
        field_info: dict[str, Any] = {}
        if hasattr(model_cls, "model_fields"):
            # Try to find field by name or alias
            field = None
            if field_name in model_cls.model_fields:
                field = model_cls.model_fields[field_name]
            else:
                # Search by alias (serialization_alias or validation_alias)
                for finfo in model_cls.model_fields.values():
                    if (
                        hasattr(finfo, "serialization_alias")
                        and finfo.serialization_alias == field_name
                    ):
                        field = finfo
                        break
                    if hasattr(finfo, "validation_alias"):
                        alias_choices = finfo.validation_alias
                        if hasattr(alias_choices, "choices") and field_name in getattr(
                            alias_choices, "choices", []
                        ):
                            field = finfo
                            break

            if field:
                field_info["description"] = field.description
                field_info["annotation"] = field.annotation

                # Handle default values properly, including enums and default_factory
                if field.default is not ...:
                    default_val = field.default
                    # If it's an enum, get its value
                    if hasattr(default_val, "value"):
                        field_info["default"] = default_val.value
                    else:
                        field_info["default"] = default_val
                elif hasattr(field, "default_factory") and field.default_factory is not None:
                    # Handle default_factory case - create instance to get defaults
                    try:
                        factory_instance = field.default_factory()  # type: ignore
                        if isinstance(factory_instance, BaseModel):
                            field_info["default"] = factory_instance.model_dump()
                        else:
                            field_info["default"] = factory_instance
                    except Exception:
                        # If factory fails, just set to None
                        pass
                    field_info["default"] = None
                else:
                    field_info["default"] = None

                field_info["examples"] = getattr(field, "examples", None)
        return field_info

    def _process_property(
        prop_name: str,
        prop_schema: dict[str, Any],
        indent: int = 0,
        model_cls: Optional[Type[BaseModel]] = None,
    ) -> list[str]:
        """Process a single property and return TOML lines with comments."""
        lines = []
        indent_str = "  " * indent

        # Get field information from Pydantic model if available
        field_info = {}
        if model_cls:
            field_info = _get_field_info(model_cls, prop_name, indent=indent)

        # Add description as comment if available
        description = field_info.get("description") or prop_schema.get("description")
        if description:
            if "This field is injected at runtime" in description:
                # Skip runtimme-generated fields.
                return []
            comment_lines = _wrap_comment(description)
            for line in comment_lines.split("\n"):
                lines.append(f"{indent_str}{line}")

        # Get the value to use - prioritize Field examples over schema examples
        value = None

        # Transform prop_schema into value expression
        optional = False
        union = False
        default = None
        example = None
        fmt_ctx = FormatterContext()

        if annotation := field_info.get("annotation", None):
            if isinstance(annotation, type):
                fmt_ctx.annotation = annotation
                if issubclass(annotation, BinarySize):
                    fmt_ctx.hint = "BinarySize"
                if issubclass(annotation, HostPortPair):
                    fmt_ctx.hint = "HostPortPair"
                if issubclass(annotation, enum.Enum):
                    if annotation.__name__ in ("AffinityPolicy",):
                        fmt_ctx.hint = "EnumByName"
                    else:
                        fmt_ctx.hint = "EnumByValue"

        match prop_schema:
            case {"anyOf": [v1, v2]}:
                if v1 == {"type": "null"}:
                    # optional value
                    value = v2
                    optional = True
                elif v2 == {"type": "null"}:
                    # optional value
                    value = v1
                    optional = True
                else:
                    # union types
                    value = [v1, v2]
                    union = True
            case {"default": v} if v != PydanticUndefined and v != ...:
                value = v
                default = v
            case _:
                value = _base_values_by_type.get(prop_schema.get("type", "string"), "?")
        if default is None:
            default = field_info.get("default", PydanticUndefined)
            if default is PydanticUndefined:
                default = None
        if (_examples := prop_schema.get("examples", PydanticUndefined)) is not PydanticUndefined:
            if _examples:
                example = _examples[0]
        if (_example := prop_schema.get("example", PydanticUndefined)) is not PydanticUndefined:
            example = _example
        if (_enum := prop_schema.get("enum", PydanticUndefined)) is not PydanticUndefined:
            example = _enum[0]

        # Format the property line
        line = indent_str
        print(
            f"{indent_str}{prop_name} ({optional=}, {union=}, {default=}, {example=}, {annotation=})",
        )
        match value:
            case None:
                # null does not exist in TOML, so just leave it as empty space
                line += f"# {prop_name} ="
            case _ if union:
                assert isinstance(value, Sequence)
                if not optional:
                    line += "## "
                line += f"{prop_name} = "
                line += _dump_toml_scalar(value[0], example, fmt_ctx)
                # Append additional unions as comments
                line += "  # | "
                line += " | ".join(_dump_toml_scalar(v, example, fmt_ctx) for v in value[1:])
            case _:
                if optional:
                    line += "## "
                # The default is likely to be None if optional.
                # Take the example value as the placeholder.
                line += f"{prop_name} = {_dump_toml_scalar(value, example if default is None else default, fmt_ctx)}"

        lines.append(line)
        return lines

    def _process_schema(
        schema_dict: dict[str, Any],
        path: Optional[list[str]] = None,
        parent_required: Optional[list[str]] = None,
        model_cls: Optional[Type[BaseModel]] = None,
    ) -> list[str]:
        """Recursively process schema and generate TOML lines."""
        if path is None:
            path = []
        if parent_required is None:
            parent_required = []

        lines = []
        properties = schema_dict.get("properties", {})
        required = schema_dict.get("required", [])

        # Group properties by type
        simple_props = {}
        object_props = {}

        for prop_name, prop_schema in properties.items():
            if "$ref" in prop_schema:
                # Resolve reference
                ref_path = prop_schema["$ref"].split("/")
                if ref_path[0] == "#" and len(ref_path) > 1:
                    resolved = schema
                    for part in ref_path[1:]:
                        resolved = resolved.get(part, {})
                    prop_schema = resolved

            prop_type = prop_schema.get("type", "")

            if (prop_type == "object" or "properties" in prop_schema) and prop_schema[
                "title"
            ] != "HostPortPair":
                object_props[prop_name] = prop_schema
            else:
                simple_props[prop_name] = prop_schema

        # Add simple properties first
        processed_simple_props = []
        for prop_name, prop_schema in simple_props.items():
            prop_lines = _process_property(
                prop_name, prop_schema, indent=len(path), model_cls=model_cls
            )
            if prop_lines:
                lines.extend(prop_lines)
                processed_simple_props.append(prop_name)

        if path == [] and processed_simple_props:
            # ref: https://github.com/toml-lang/toml/issues/984
            warnings.append(
                "The configuration schema CANNOT have simple fields in the root "
                "without any section header according to the TOML specification. "
                "Also, optional sections should be defined non-optional with explicit default factory. "
                f"Please move or fix these fields/sections: {', '.join(simple_props.keys())}. "
            )

        # Add object properties as sections
        for prop_name, prop_schema in object_props.items():
            indent_str = "  " * len(path)

            if lines and lines[-1].strip():  # Add blank line before section
                lines.append("")

            # Add section comment
            if "description" in prop_schema:
                comment_lines = _wrap_comment(prop_schema["description"], prefix=indent_str)
                lines.extend(comment_lines.split("\n"))

            # Add section header
            section_path = path + [prop_name]
            section_header = f"{indent_str}[{'.'.join(section_path)}]"
            lines.append(section_header)
            print(section_header)

            # Process nested properties - try to get nested model class
            nested_model_cls = None
            if model_cls and hasattr(model_cls, "model_fields"):
                field_info = _get_field_info(model_cls, prop_name, indent=len(path))
                if field_info:
                    # First try to find the field by name or alias
                    field = None
                    if prop_name in model_cls.model_fields:
                        field = model_cls.model_fields[prop_name]
                    else:
                        # Search by alias
                        for finfo in model_cls.model_fields.values():
                            if (
                                hasattr(finfo, "serialization_alias")
                                and finfo.serialization_alias == prop_name
                            ):
                                field = finfo
                                break

                    if field:
                        if hasattr(field, "annotation") and hasattr(field.annotation, "__origin__"):
                            # Handle generic types
                            args = getattr(field.annotation, "__args__", ())
                            if args and hasattr(args[0], "model_fields"):
                                nested_model_cls = args[0]
                        elif hasattr(field.annotation, "model_fields"):
                            nested_model_cls = field.annotation

            nested_lines = _process_schema(
                prop_schema, path=section_path, parent_required=required, model_cls=nested_model_cls
            )
            lines.extend(nested_lines)

        return lines

    # Process the root schema
    lines = _process_schema(schema, model_cls=model_class)
    for msg in warnings:
        log.warning(msg)
    return "\n".join(lines)


def generate_sample_config_file(
    model_class: Type[BaseModel], output_path: str, header_comment: Optional[str] = None
) -> None:
    """
    Generate a sample TOML configuration file and write it to disk.

    Args:
        model_class: The Pydantic model class to generate configuration for
        output_path: Path where the configuration file will be written
        header_comment: Optional header comment to add at the top of the file
    """
    try:
        config_content = _generate_sample_config(model_class)
    except Exception:
        log.exception("Error while generating sample config schema")
    with open(output_path, "w") as f:
        if header_comment:
            f.write(_wrap_comment(header_comment))
            f.write("\n\n")
        f.write(config_content)
        f.write("\n")



"""Leader tasks subpackage for periodic and event-driven tasks."""

from .base import PeriodicTask
from .event_task import EventProducerTask, EventTaskSpec
from .leader_cron import LeaderCron

__all__ = [
    "EventProducerTask",
    "EventTaskSpec",
    "LeaderCron",
    "PeriodicTask",
]



from dataclasses import asdict, dataclass
from typing import Any, Mapping, Optional, override

from .user_event import UserEvent


@dataclass
class UserBgtaskUpdatedEvent(UserEvent):
    task_id: str
    message: str
    current_progress: float
    total_progress: float

    @override
    def event_name(self) -> Optional[str]:
        return "bgtask_updated"

    @override
    def retry_count(self) -> Optional[int]:
        return 5

    @override
    def user_event_mapping(self) -> Mapping[str, Any]:
        return asdict(self)

    @override
    def is_close_event(self) -> bool:
        return False


@dataclass
class UserBgtaskDoneEvent(UserEvent):
    task_id: str
    message: str

    @override
    def event_name(self) -> Optional[str]:
        return "bgtask_done"

    @override
    def retry_count(self) -> Optional[int]:
        return None

    @override
    def user_event_mapping(self) -> Mapping[str, Any]:
        return asdict(self)

    @override
    def is_close_event(self) -> bool:
        return True


@dataclass
class UserBgtaskPartialSuccessEvent(UserEvent):
    task_id: str
    message: str
    errors: list[str]

    @override
    def event_name(self) -> Optional[str]:
        return "bgtask_done"

    @override
    def retry_count(self) -> Optional[int]:
        return None

    @override
    def user_event_mapping(self) -> Mapping[str, Any]:
        return asdict(self)

    @override
    def is_close_event(self) -> bool:
        return True


@dataclass
class UserBgtaskCancelledEvent(UserEvent):
    task_id: str
    message: str

    @override
    def event_name(self) -> Optional[str]:
        return "bgtask_cancelled"

    @override
    def retry_count(self) -> Optional[int]:
        return None

    @override
    def user_event_mapping(self) -> Mapping[str, Any]:
        return asdict(self)

    @override
    def is_close_event(self) -> bool:
        return True


@dataclass
class UserBgtaskFailedEvent(UserEvent):
    task_id: str
    message: str

    @override
    def event_name(self) -> Optional[str]:
        return "bgtask_failed"

    @override
    def retry_count(self) -> Optional[int]:
        return None

    @override
    def user_event_mapping(self) -> Mapping[str, Any]:
        return asdict(self)

    @override
    def is_close_event(self) -> bool:
        return True
